"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import logging
import uuid
from collections import defaultdict
from datetime import datetime
from re import sub
from typing import Any, Dict, List
from uuid import UUID, uuid4

import pytest

import utils
from inmanta import config, const, data
from inmanta.agent.agent import Agent
from inmanta.const import ResourceAction, ResourceState
from inmanta.server import SLICE_AGENT_MANAGER, SLICE_ORCHESTRATION, SLICE_RESOURCE
from inmanta.server.services.orchestrationservice import OrchestrationService
from inmanta.server.services.resourceservice import ResourceService
from inmanta.util import get_compiler_version


class MultiVersionSetup(object):
    """
      create scenarios by describing the history of a resource, from newest to oldest state

      V = void, nothing
      A - available/skipped/unavailable
      E - error
      D - deployed
      d - deploying
      p - processing events
      S - skipped for undefined
      U - undefined
    """

    def __init__(self):
        self.firstversion: int = 100
        self.versions: List[List[Dict[str, Any]]] = [[] for _ in range(100)]
        self.states: Dict[str, ResourceState] = {}
        self.results: Dict[str, List[str]] = defaultdict(lambda: [])
        self.sid = uuid4()

    def get_version(self, v: int) -> List[Dict[str, Any]]:
        return self.versions[v]

    def expand_code(self, code: str) -> ResourceState:
        if code == "A":
            return ResourceState.available

        if code == "E":
            return ResourceState.failed

        if code == "D":
            return ResourceState.deployed

        if code == "d":
            return ResourceState.deploying

        if code == "p":
            return ResourceState.processing_events

        if code == "S":
            return ResourceState.skipped_for_undefined

        if code == "U":
            return ResourceState.undefined

        assert False

    def make_resource(
        self, name: str, value: str, version: int, agent: str = "agent1", requires: List[str] = [], send_event: bool = False
    ) -> str:
        """
            requires: list of resource identifiers
        """
        id = "test::Resource[%s,key=%s],v=%d" % (agent, name, version)
        res = {
            "key": value,
            "id": id,
            "send_event": send_event,
            "purged": False,
            "requires": ["%s,v=%d" % (r, version) for r in requires],
        }
        self.get_version(version).append(res)
        return id

    def add_resource(
        self, name: str, scenario: str, increment: bool, agent="agent1", requires=[], send_event: bool = False
    ) -> str:
        v = self.firstversion
        rid = "test::Resource[%s,key=%s]" % (agent, name)

        if increment:
            self.results[agent].append(rid)

        for step in scenario.split():
            v -= 1
            code = step[0]
            if code == "V":
                continue
            value = step[1:]
            rvid = self.make_resource(name, value, v, agent, requires, send_event)
            self.states[rvid] = self.expand_code(code)
        return rid

    async def setup_agent(self, server, environment):

        agentmanager = server.get_slice(SLICE_AGENT_MANAGER)

        endpoints = list(self.results.keys())

        config.Config.set("config", "agent-deploy-interval", "0")
        config.Config.set("config", "agent-repair-interval", "0")

        a = Agent(hostname="node1", environment=environment, agent_map={e: "localhost" for e in endpoints}, code_loader=False)
        for e in endpoints:
            await a.add_end_point_name(e)
        await a.start()
        await utils.retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

        return a

    async def setup(
        self, serverdirect: OrchestrationService, resource_service: ResourceService, env: data.Environment, sid: UUID
    ):
        for version in range(0, len(self.versions)):
            # allocate a bunch of versions!
            v = await env.get_next_version()
            assert v == version + 1
            if self.versions[version]:
                res = await serverdirect.put_version(
                    env=env,
                    version=version,
                    resources=self.versions[version],
                    unknowns=[],
                    version_info={},
                    resource_state={},
                    compiler_version=get_compiler_version(),
                )
                assert res == 200

                result, _ = await serverdirect.release_version(env, version, False)
                assert result == 200
        for rid, state in self.states.items():
            # Start the deploy
            action_id = uuid.uuid4()
            now = datetime.now()
            if state == const.ResourceState.available:
                # initial state can not be set
                continue
            if state not in const.TRANSIENT_STATES:
                finished = now
            else:
                finished = None
            result = await resource_service.resource_action_update(
                env,
                [rid],
                action_id,
                ResourceAction.deploy,
                now,
                finished,
                status=state,
                messages=[],
                changes={},
                change=None,
                send_events=False,
            )
            assert result == 200

        allresources = {}

        for agent, results in self.results.items():
            result, payload = await resource_service.get_resources_for_agent(
                env, agent, version=None, incremental_deploy=True, sid=sid
            )

            assert sorted([x["resource_id"] for x in payload["resources"]]) == sorted(results)
            allresources.update({r["resource_id"]: r for r in payload["resources"]})

        return allresources


@pytest.mark.asyncio
async def test_deploy(server, agent: Agent, environment, caplog):
    """
        Test basic deploy mechanism mocking
    """
    with caplog.at_level(logging.WARNING):
        # acquire raw server
        orchestration_service = server.get_slice(SLICE_ORCHESTRATION)
        resource_service = server.get_slice(SLICE_RESOURCE)
        sid = agent.sessionid

        # acquire env object
        env = await data.Environment.get_by_id(uuid.UUID(environment))

        version = await env.get_next_version()

        def make_resources(version):
            return [
                {
                    "key": "key1",
                    "id": "test::Resource[agent1,key=key1],v=%d" % version,
                    "send_event": False,
                    "purged": False,
                    "requires": [],
                },
                {
                    "key": "key2",
                    "id": "test::Resource[agent1,key=key2],v=%d" % version,
                    "send_event": False,
                    "requires": ["test::Resource[agent1,key=key2],v=%d" % version],
                    "purged": False,
                },
                {
                    "key": "key3",
                    "id": "test::Resource[agent1,key=key3],v=%d" % version,
                    "send_event": False,
                    "requires": ["test::Resource[agent2,key=key4],v=%d" % version],
                    "purged": True,
                },
                {"key": "key4", "id": "test::Resource[agent2,key=key4],v=%d" % version, "send_event": False, "requires": []},
            ]

        resources = make_resources(version)
        res = await orchestration_service.put_version(
            env=env,
            version=version,
            resources=resources,
            unknowns=[],
            version_info={},
            resource_state={},
            compiler_version=get_compiler_version(),
        )
        assert res == 200

        result, _ = await orchestration_service.release_version(env, version, False)
        assert result == 200

        resource_ids = [x["id"] for x in resources]

        # Start the deploy
        action_id = uuid.uuid4()
        now = datetime.now()
        result = await resource_service.resource_action_update(
            env,
            resource_ids,
            action_id,
            const.ResourceAction.deploy,
            now,
            now,
            status=ResourceState.deployed,
            messages=[],
            changes={},
            change=None,
            send_events=False,
        )
        assert result == 200

        result, payload = await orchestration_service.get_version(env, version)
        assert result == 200
        assert payload["model"].done == len(resources)

        # second, identical check_version
        v2 = await env.get_next_version()
        resources = make_resources(v2)
        res = await orchestration_service.put_version(
            env=env,
            version=v2,
            resources=resources,
            unknowns=[],
            version_info={},
            resource_state={},
            compiler_version=get_compiler_version(),
        )
        assert res == 200

        result, payload = await resource_service.get_resources_for_agent(
            env, "agent1", version=None, incremental_deploy=True, sid=sid
        )
        assert len(payload["resources"]) == 0

        # Cannot request increment for specific version
        result, _ = await resource_service.get_resources_for_agent(
            env, "agent1", version=version, incremental_deploy=True, sid=sid
        )
        assert result == 500
        result, _ = await resource_service.get_resources_for_agent(env, "agent1", version=v2, incremental_deploy=True, sid=sid)
        assert result == 500

    for record in caplog.records:
        assert record.levelname != "WARNING"


def strip_version(v):
    return sub(",v=[0-9]+", "", v)


@pytest.mark.asyncio
async def test_deploy_scenarios(server, agent: Agent, environment, caplog):
    with caplog.at_level(logging.WARNING):
        # acquire raw server
        orchestration_service = server.get_slice(SLICE_ORCHESTRATION)
        resource_service = server.get_slice(SLICE_RESOURCE)
        sid = agent.sessionid

        # acquire env object
        env = await data.Environment.get_by_id(uuid.UUID(environment))

        setup = MultiVersionSetup()

        setup.add_resource("R1", "A1 A1 V1 D1", True)
        setup.add_resource("R2", "A1 E1 D1", True)
        setup.add_resource("R3", "A1 E1 E1 A1", True)
        setup.add_resource("R4", "A1 D1", False)
        setup.add_resource("R5", "A1 A2 D1", False)
        setup.add_resource("R6", "A1 D2", True)
        setup.add_resource("R7", "A1 D2 D1", True)
        setup.add_resource("R8", "D1 A1 E1 D1", False)
        setup.add_resource("R9", "A1 E2 D1", True)
        setup.add_resource("R10", "A1 A1 D1", False)
        setup.add_resource("R13", "A1 A1 A1 A1 A1", True)
        setup.add_resource("R14", "A1 A1 d1 D1", True)
        setup.add_resource("R15", "A1 A1 p1 D1", True)
        setup.add_resource("R16", "S1 A1", False)
        setup.add_resource("R17", "A1 S1 A1", True)
        setup.add_resource("R18", "D1 S1 A1", False)
        setup.add_resource("R19", "U1 A1", False)
        setup.add_resource("R20", "U1 D1", False)
        setup.add_resource("R21", "A1 U1", True)

        await setup.setup(orchestration_service, resource_service, env, sid)

    for record in caplog.records:
        assert record.levelname != "WARNING"


@pytest.mark.asyncio
async def test_deploy_scenarios_removed_req_by_increment(server, agent: Agent, environment, caplog):
    with caplog.at_level(logging.WARNING):
        # acquire raw server
        orchestration_service = server.get_slice(SLICE_ORCHESTRATION)
        resource_service = server.get_slice(SLICE_RESOURCE)
        sid = agent.sessionid

        # acquire env object
        env = await data.Environment.get_by_id(uuid.UUID(environment))

        setup = MultiVersionSetup()

        id1 = setup.add_resource("R1", "A1 D1", False)
        id2 = setup.add_resource("R2", "A1 D2", True, requires=[id1])

        resources = await setup.setup(orchestration_service, resource_service, env, sid)
        assert not resources[id2]["attributes"]["requires"]

    for record in caplog.records:
        assert record.levelname != "WARNING"


@pytest.mark.asyncio
async def test_deploy_scenarios_removed_req_by_increment2(server, environment, caplog):
    with caplog.at_level(logging.WARNING):
        # acquire raw server
        orchestration_service = server.get_slice(SLICE_ORCHESTRATION)
        resource_service = server.get_slice(SLICE_RESOURCE)

        # acquire env object
        env = await data.Environment.get_by_id(uuid.UUID(environment))

        setup = MultiVersionSetup()

        id1 = setup.add_resource("R1", "A1 D1", False)
        id3 = setup.add_resource("R3", "A1 D2", True)
        id4 = setup.add_resource("R4", "A1 D2", True, agent="agent2")
        id2 = setup.add_resource("R2", "A1 D2", True, requires=[id1, id3, id4])

        agent = await setup.setup_agent(server, environment)

        sid = agent.sessionid

        try:
            resources = await setup.setup(orchestration_service, resource_service, env, sid)
            print(sorted(resources[id2]["attributes"]["requires"]))
            print(sorted(sorted([id3, id4])))
            assert sorted([strip_version(r) for r in resources[id2]["attributes"]["requires"]]) == sorted([id3, id4])

        finally:
            await agent.stop()
    for record in caplog.records:
        # gets some logs when setting up agent
        assert record.levelname != "WARNING" or record.name == "inmanta.config"


@pytest.mark.asyncio
async def test_deploy_scenarios_added_by_send_event(server, agent: Agent, environment, caplog):
    with caplog.at_level(logging.WARNING):
        # acquire raw server
        orchestration_service = server.get_slice(SLICE_ORCHESTRATION)
        resource_service = server.get_slice(SLICE_RESOURCE)
        sid = agent.sessionid

        # acquire env object
        env = await data.Environment.get_by_id(uuid.UUID(environment))

        setup = MultiVersionSetup()

        id1 = setup.add_resource("R1", "A1 D2", True, send_event=True)
        id2 = setup.add_resource("R2", "A1 D1", True, requires=[id1])
        id3 = setup.add_resource("R3", "A1 D1", True, requires=[id1], send_event=True)
        setup.add_resource("R4", "A1 D1", True, requires=[id3])
        setup.add_resource("R5", "A1 D1", False, requires=[id2])

        await setup.setup(orchestration_service, resource_service, env, sid)

    for record in caplog.records:
        assert record.levelname != "WARNING"


@pytest.mark.asyncio
async def test_deploy_scenarios_added_by_send_event_cad(server, agent: Agent, environment, caplog):
    # ensure CAD does not change send_event
    with caplog.at_level(logging.WARNING):
        # acquire raw server
        orchestration_service = server.get_slice(SLICE_ORCHESTRATION)
        resource_service = server.get_slice(SLICE_RESOURCE)
        sid = agent.sessionid

        # acquire env object
        env = await data.Environment.get_by_id(uuid.UUID(environment))

        setup = MultiVersionSetup()

        id1 = setup.add_resource("R1", "A1 D2", True, send_event=False)
        id2 = setup.add_resource("R2", "A1 D1", False, requires=[id1])
        id3 = setup.add_resource("R3", "A1 D1", False, requires=[id1], send_event=True)
        setup.add_resource("R4", "A1 D1", False, requires=[id3])
        setup.add_resource("R5", "A1 D1", False, requires=[id2])

        setup.add_resource("R6", "A1 D1", False, requires=[id1], agent="agent2")
        await setup.setup(orchestration_service, resource_service, env, sid)

    for record in caplog.records:
        assert record.levelname != "WARNING"



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import pytest

import inmanta.compiler as compiler
from inmanta.ast import DuplicateException, RuntimeException, TypingException
from inmanta.execute.runtime import Instance


def test_dict(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
a = "a"
b = { "a" : a, "b" : "b", "c" : 3}
"""
    )

    (_, root) = compiler.do_compile()

    scope = root.get_child("__config__").scope
    b = scope.lookup("b").get_value()
    assert b["a"] == "a"
    assert b["b"] == "b"
    assert b["c"] == 3


def test_dict_collide(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
a = "a"
b = { "a" : a, "a" : "b", "c" : 3}
"""
    )

    with pytest.raises(DuplicateException):
        compiler.do_compile()


def test_dict_attr(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Foo:
  dict bar
  dict foo = {}
  dict blah = {"a":"a"}
end

implement Foo using std::none

a=Foo(bar={})
b=Foo(bar={"a":z})
c=Foo(bar={}, blah={"z":"y"})
z=5
"""
    )

    (_, root) = compiler.do_compile()

    scope = root.get_child("__config__").scope

    def map_assert(in_dict, expected):
        for (ek, ev), (k, v) in zip(expected.items(), in_dict.items()):
            assert ek == k
            assert ev == v

    def validate(var, bar, foo, blah):
        e = scope.lookup(var).get_value()
        map_assert(e.get_attribute("bar").get_value(), bar)
        map_assert(e.get_attribute("foo").get_value(), foo)
        map_assert(e.get_attribute("blah").get_value(), blah)

    validate("a", {}, {}, {"a": "a"})
    validate("b", {"a": 5}, {}, {"a": "a"})

    validate("c", {}, {}, {"z": "y"})


def test_dict_attr_type_error(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Foo:
  dict bar
  dict foo = {}
  dict blah = {"a":"a"}
end

implement Foo using std::none

a=Foo(bar=b)
b=Foo(bar={"a":"A"})
"""
    )
    with pytest.raises(RuntimeException):
        compiler.do_compile()


def test_611_dict_access(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
a = "a"
b = { "a" : a, "b" : "b", "c" : 3}
c=b[a]
d=b["c"]
"""
    )

    (_, root) = compiler.do_compile()

    scope = root.get_child("__config__").scope
    assert scope.lookup("c").get_value() == "a"
    assert scope.lookup("d").get_value() == 3


def test_632_dict_access_2(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
b = { "a" : {"b":"c"}}
c=b["a"]["b"]
"""
    )

    (_, root) = compiler.do_compile()

    scope = root.get_child("__config__").scope
    assert scope.lookup("c").get_value() == "c"


def test_632_dict_access_3(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
b = { "a" : "b"}
c=b["a"]["b"]
"""
    )

    with pytest.raises(TypingException):
        compiler.do_compile()


def test_673_in_dict(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    dict attributes
end

implementation test for Test:

end

implement Test using test when "foo" in self.attributes

Test(attributes={"foo": 42})
"""
    )
    compiler.do_compile()


def test_bad_map_lookup(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
        b = {"c" : 3}
        c=b["a"]
        """,
        "key a not found in dict, options are [c] (reported in c = b['a'] ({dir}/main.cf:3))",
    )


def test_1168_const_dict(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """typedef test as string matching std::contains({"1234":"xxx"}, self)

entity X:
    test x
end

X(x="1234")

implement X using std::none
"""
    )

    compiler.do_compile()


def test_1168_const_dict_failure(snippetcompiler):
    snippetcompiler.setup_for_error(
        """typedef test as string matching std::contains({"1234":z}, self)
z = "1234"

entity X:
    test x
end

X(x="1234")

implement X using std::none
""",
        """Could not set attribute `x` on instance `__config__::X (instantiated at {dir}/main.cf:8)` (reported in Construct(X) ({dir}/main.cf:8))
caused by:
  Unable to resolve `z`: a type constraint can not reference variables. (reported in z ({dir}/main.cf:1:55))""",  # noqa: E501,
    )


def test_constructor_kwargs(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    number n
    number m
    string str
end

implement Test using std::none

dct = { "config": {"n": 42, "m": 0 } }

x = Test(**dct["config"], str = "Hello World!")
"""
    )
    (_, root) = compiler.do_compile()
    scope = root.get_child("__config__").scope

    instance: Instance = scope.lookup("x").get_value()
    assert instance.get_attribute("n").get_value() == 42
    assert instance.get_attribute("m").get_value() == 0
    assert instance.get_attribute("str").get_value() == "Hello World!"


@pytest.mark.parametrize("override", [True, False])
def test_2003_constructor_kwargs_default(snippetcompiler, override: bool):
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    number v = 0
end

implement Test using std::none

values = {%s}
t = Test(
    **values
)
        """
        % ("'v': 10" if override else ""),
    )
    (_, root) = compiler.do_compile()
    scope = root.get_child("__config__").scope

    instance: Instance = scope.lookup("t").get_value()
    assert instance.get_attribute("v").get_value() == (10 if override else 0)


def test_constructor_kwargs_double_set(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
entity A:
    int a
end

v = {"a": 4}
A(a = 3, **v)
        """,
        "The attribute a is set twice in the constructor call of A. (reported in Construct(A) ({dir}/main.cf:7))",
    )


def test_constructor_kwargs_index_match(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    number n
    number m
    string str
end

index Test(n, m)

implement Test using std::none

dct = { "config": {"n": 42, "m": 0 } }

x = Test(n = 42, m = 0, str = "Hello World!")
y = Test(**dct["config"], str = "Hello World!")
"""
    )
    (_, root) = compiler.do_compile()
    scope = root.get_child("__config__").scope

    x: Instance = scope.lookup("x").get_value()
    y: Instance = scope.lookup("y").get_value()
    assert x is y


def test_indexlookup_kwargs(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    number n
    number m
    string str
end

index Test(n, m)

implement Test using std::none

dct = {"m": 0}
x = Test(n = 42, m = 0, str = "Hello World!")

y = Test[n = 42, **dct]
    """
    )
    (_, root) = compiler.do_compile()
    scope = root.get_child("__config__").scope

    x: Instance = scope.lookup("x").get_value()
    y: Instance = scope.lookup("y").get_value()
    assert x is y


def test_short_indexlookup_kwargs(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Collection:
end

implement Collection using std::none

entity Test:
    number n
    number m
    string str
end

Collection.tests [0:] -- Test.collection [1]
index Test(collection, n, m)

implement Test using std::none

dct = {"m": 0}
c = Collection()

x = Test(collection = c, n = 42, m = 0, str = "Hello World!")
y = Test(collection = c, n = 0, m = 0, str = "Hello World!")

z = c.tests[n = 42, **dct]
    """
    )
    (_, root) = compiler.do_compile()
    scope = root.get_child("__config__").scope

    x: Instance = scope.lookup("x").get_value()
    z: Instance = scope.lookup("z").get_value()
    assert x is z



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import base64
import typing
from typing import TypeVar

import pytest

from inmanta.agent.handler import ResourceHandler
from inmanta.protocol import SessionClient, VersionMatch, common
from test_protocol import make_random_file

T = TypeVar("T")


class MockSessionClient(SessionClient):
    def __init__(self, return_code, content):
        self._version_match = VersionMatch.highest
        self.return_code = return_code
        self.content = content
        pass

    def get_file(self, hash_id):
        content = b""
        if self.return_code != 404:
            content = base64.b64encode(self.content)
        return common.Result(self.return_code, result={"content": content})


class MockGetFileResourceHandler(ResourceHandler):
    def __init__(self, client):
        self._client = client

    def run_sync(self, func: typing.Callable[[], T]) -> T:
        return func()


def test_get_file_corrupted():
    (hash, content, body) = make_random_file()
    client = MockSessionClient(200, b"corrupted_file")
    resource_handler = MockGetFileResourceHandler(client)

    with pytest.raises(Exception):
        resource_handler.get_file(hash)


def test_get_file_success():
    (hash, content, body) = make_random_file()
    client = MockSessionClient(200, content)
    resource_handler = MockGetFileResourceHandler(client)

    result = resource_handler.get_file(hash)
    assert content == result


def test_get_file_not_found():
    client = MockSessionClient(404, None)
    resource_handler = MockGetFileResourceHandler(client)
    result = resource_handler.get_file("hash")
    assert result is None



"""
    Copyright 2020 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import inspect
import json
import typing
import uuid
from datetime import datetime
from enum import Enum
from typing import Callable, Dict, List, Optional, Type, Union

import typing_inspect  # type: ignore
from pydantic.main import BaseModel
from pydantic.networks import AnyUrl
from pydantic.schema import model_schema

from inmanta import types, util
from inmanta.protocol.common import ArgOption, MethodProperties, UrlMethod
from inmanta.protocol.openapi.model import (
    Components,
    Header,
    Info,
    MediaType,
    OpenAPI,
    Operation,
    Parameter,
    ParameterType,
    PathItem,
    Reference,
    RequestBody,
    Response,
    Schema,
    SchemaBase,
    Server,
)
from inmanta.server import config
from inmanta.server.extensions import FeatureManager


def openapi_json_encoder(o) -> Union[Dict, str, List]:
    if isinstance(o, BaseModel):
        return o.dict(by_alias=True, exclude_none=True)
    return util.custom_json_encoder(o)


class OpenApiConverter:
    """
        Extracts API information for the OpenAPI definition from the server
    """

    def __init__(self, global_url_map: Dict[str, Dict[str, UrlMethod]], feature_manager: FeatureManager):
        self.global_url_map = global_url_map
        self.feature_manager = feature_manager
        self.type_converter = OpenApiTypeConverter()
        self.arg_option_handler = ArgOptionHandler(self.type_converter)

    def _collect_server_information(self) -> List[Server]:
        bind_port = config.get_bind_port()
        server_address = config.server_address.get()
        return [
            Server(url=AnyUrl(url=f"http://{server_address}:{bind_port}/", scheme="http", host=server_address, port=bind_port))
        ]

    def _get_inmanta_version(self) -> Optional[str]:
        metadata = self.feature_manager.get_product_metadata()
        return metadata["version"]

    def generate_openapi_definition(self) -> OpenAPI:
        version = self._get_inmanta_version()
        info = Info(title="Inmanta Service Orchestrator", version=version if version else "")
        servers = self._collect_server_information()
        paths = {}
        for path, methods in self.global_url_map.items():
            api_methods = self._filter_api_methods(methods)
            if len(api_methods) > 0:
                path_in_openapi_format = self._format_path(path)
                path_item = self._extract_operations_from_methods(api_methods, path_in_openapi_format)
                paths[path_in_openapi_format] = path_item
        return OpenAPI(openapi="3.0.2", info=info, paths=paths, servers=servers, components=self.type_converter.components)

    def _filter_api_methods(self, methods: Dict[str, UrlMethod]) -> Dict[str, UrlMethod]:
        return {
            method_name: url_method
            for method_name, url_method in methods.items()
            if "api" in url_method.properties.client_types
        }

    def _format_path(self, path: str) -> str:
        return path.replace("(?P<", "{").replace(">[^/]+)", "}")

    def _extract_operations_from_methods(self, api_methods: Dict[str, UrlMethod], path: str) -> PathItem:
        path_item = PathItem()
        for http_method_name, url_method in api_methods.items():
            operation_handler = OperationHandler(self.type_converter, self.arg_option_handler)
            operation = operation_handler.handle_method(url_method, path)
            path_item.__setattr__(http_method_name.lower(), operation)
        return path_item

    def generate_openapi_json(self) -> str:
        openapi = self.generate_openapi_definition()

        return json.dumps(openapi, default=openapi_json_encoder)

    def generate_swagger_html(self) -> str:
        return self.get_swagger_html(self.generate_openapi_json())

    def get_swagger_html(self, openapi_spec: str) -> str:
        template = f"""
        <!DOCTYPE html>
    <html lang="en">
    <head>
        <meta charset="UTF-8">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
        <script src="//unpkg.com/swagger-ui-dist@^3.25.0/swagger-ui-bundle.js"></script>
        <link rel="stylesheet" href="//unpkg.com/swagger-ui-dist@^3.25.0/swagger-ui.css" />
        <title>Inmanta Service Orchestrator API</title>
    </head>
    <body>
        <div id="swagger-ui"></div>
        <script>
            window.onload = function() {{
              SwaggerUIBundle({{
                spec: {openapi_spec},
                dom_id: '#swagger-ui',
                presets: [
                  SwaggerUIBundle.presets.apis
                ],
              }})
            }}
        </script>
    </body>
    </html>
        """

        return template


class OpenApiTypeConverter:
    """
        Lookup for OpenAPI types corresponding to python types
    """

    components = Components(schemas={})

    python_to_openapi_types = {
        bool: Schema(type="boolean"),
        int: Schema(type="integer"),
        str: Schema(type="string"),
        tuple: Schema(type="array", items=Schema()),
        list: Schema(type="array", items=Schema()),
        dict: Schema(type="object"),
        float: Schema(type="number", format="float"),
        bytes: Schema(type="string", format="binary"),
        datetime: Schema(type="string", format="date-time"),
        uuid.UUID: Schema(type="string", format="uuid"),
        typing.Any: Schema(),
        types.StrictNonIntBool: Schema(type="boolean"),
    }

    def get_openapi_type_of_parameter(self, parameter_type: inspect.Parameter) -> Schema:
        type_annotation = parameter_type.annotation
        return self.get_openapi_type(type_annotation)

    def _is_none_type(self, type_annotation: Type) -> bool:
        return inspect.isclass(type_annotation) and issubclass(type_annotation, type(None))

    def _handle_union_type(self, type_annotation: Type) -> Schema:
        # An Optional is always a Union
        type_args = typing_inspect.get_args(type_annotation, evaluate=True)
        openapi_types = [self.get_openapi_type(type_arg) for type_arg in type_args if not self._is_none_type(type_arg)]
        none_type_in_type_args = len(openapi_types) < len(type_args)
        if none_type_in_type_args:
            if len(openapi_types) == 1:
                openapi_type = openapi_types[0].copy(deep=True)
                # An Optional in the OpenAPI Schema is nullable
                openapi_type.nullable = True
                return openapi_type
            return Schema(anyOf=openapi_types, nullable=True)
        # A Union type can be expressed as a schema that matches any of the type arguments
        return Schema(anyOf=openapi_types)

    def _handle_dictionary(self, type_annotation: Type) -> Schema:
        type_args = typing_inspect.get_args(type_annotation, evaluate=True)
        return Schema(type="object", additionalProperties=self.get_openapi_type(type_args[1]))

    def _handle_pydantic_model(self, type_annotation: Type) -> Schema:
        # JsonSchema stores the model (and sub-model) definitions at #/definitions,
        # but OpenAPI requires them to be placed at "#/components/schemas/"
        # The ref_prefix changes the references, but the actual schemas are still at #/definitions
        schema = model_schema(type_annotation, by_alias=True, ref_prefix="#/components/schemas/")
        if "definitions" in schema.keys():
            definitions = schema.pop("definitions")
            if self.components.schemas is not None:
                self.components.schemas.update(definitions)
        return Schema(**schema)

    def _handle_enums(self, type_annotation: Type) -> Schema:
        enum_keys = [name for name in type_annotation.__members__.keys()]
        return Schema(type="string", enum=enum_keys)

    def _handle_list(self, type_annotation: Type) -> Schema:
        # Type argument is always present, see protocol.common.MethodProperties._validate_type_arg()
        list_member_type = typing_inspect.get_args(type_annotation, evaluate=True)
        return Schema(type="array", items=self.get_openapi_type(list_member_type[0]))

    def get_openapi_type(self, type_annotation: Type) -> Schema:
        type_origin = typing_inspect.get_origin(type_annotation)
        if typing_inspect.is_union_type(type_annotation):
            return self._handle_union_type(type_annotation)
        elif inspect.isclass(type_annotation) and issubclass(type_annotation, BaseModel):
            return self._handle_pydantic_model(type_annotation)
        elif inspect.isclass(type_annotation) and issubclass(type_annotation, Enum):
            return self._handle_enums(type_annotation)
        elif inspect.isclass(type_origin) and issubclass(type_origin, typing.Mapping):
            return self._handle_dictionary(type_annotation)
        elif inspect.isclass(type_origin) and issubclass(type_origin, typing.Sequence):
            return self._handle_list(type_annotation)
        # Fallback to primitive types
        return self.python_to_openapi_types.get(type_annotation, Schema(type="object"))


class ArgOptionHandler:
    """
        Extracts header, response header and path parameter information from ArgOptions
    """

    def __init__(self, type_converter: OpenApiTypeConverter):
        self.type_converter = type_converter

    def extract_parameters_from_arg_options(
        self, method_properties: MethodProperties, function_parameters: Dict[str, inspect.Parameter]
    ) -> List[Parameter]:
        result: List[Parameter] = []
        for option_name, option in method_properties.arg_options.items():
            param = function_parameters[option_name]
            param_schema = self.type_converter.get_openapi_type_of_parameter(param)
            param_description = method_properties.get_description_for_param(option_name)
            if option.header:
                result.append(
                    Parameter(in_=ParameterType.header, name=option.header, schema_=param_schema, description=param_description)
                )
        return result

    def extract_response_headers_from_arg_options(
        self, arg_options: Dict[str, ArgOption]
    ) -> Optional[Dict[str, Union[Header, Reference]]]:
        headers: Dict[str, Union[Header, Reference]] = {}
        for option_name, option in arg_options.items():
            if option.header and option.reply_header:
                headers[option.header] = Header(description=option.header, schema=Schema(type="string"))
        return headers if headers else None


class FunctionParameterHandler:
    """
        Creates OpenAPI Parameters and RequestBody items based on the handler function parameters
    """

    def __init__(
        self,
        type_converter: OpenApiTypeConverter,
        arg_option_handler: ArgOptionHandler,
        path: str,
        method_properties: MethodProperties,
    ):
        self.type_converter = type_converter
        self.arg_option_handler = arg_option_handler
        self.path = path
        self.method_properties = method_properties

        # Get the parameters of the handler function
        self.all_params_dct: Dict[str, inspect.Parameter] = self._extract_function_parameters(method_properties.function)
        self.path_params: Dict[str, inspect.Parameter] = {}
        self.header_params: Dict[str, inspect.Parameter] = {}
        self.non_path_and_non_header_params: Dict[str, inspect.Parameter] = {}

        for param_name, param in self.all_params_dct.items():
            if f"{{{param_name}}}" in self.path:
                self.path_params[param_name] = param
            elif (
                param_name in self.method_properties.arg_options.keys()
                and self.method_properties.arg_options[param_name].header is not None
            ):
                self.header_params[param_name] = param
            else:
                self.non_path_and_non_header_params[param_name] = param

    # Parameters

    def _extract_function_parameters(self, url_method_function: Callable) -> Dict[str, inspect.Parameter]:
        function_parameters = {
            parameter_name: parameter_type
            for parameter_name, parameter_type in inspect.signature(url_method_function).parameters.items()
        }
        return function_parameters

    def get_parameters(self) -> List[Parameter]:
        result = self._convert_header_and_path_params()
        if self.method_properties.operation not in ["POST", "PUT", "PATCH"]:
            result.extend(self._convert_function_params_to_query_params())
        return result

    def _convert_header_and_path_params(self) -> List[Parameter]:
        arg_options_params: List[Parameter] = self.arg_option_handler.extract_parameters_from_arg_options(
            self.method_properties, self.all_params_dct
        )
        path_params: List[Parameter] = self._convert_path_params_to_openapi()
        return arg_options_params + path_params

    def _convert_path_params_to_openapi(self) -> List[Parameter]:
        parameters: List[Union[Parameter, Reference]] = []
        for parameter_name, parameter_type in self.path_params.items():
            type_description = self.type_converter.get_openapi_type_of_parameter(parameter_type)
            param_description = self.method_properties.get_description_for_param(parameter_name)
            parameters.append(
                Parameter(
                    name=parameter_name,
                    in_=ParameterType.path,
                    required=True,
                    schema_=type_description,
                    description=param_description,
                )
            )
        return parameters

    def _convert_function_params_to_query_params(self) -> List[Parameter]:
        parameters: List[Parameter] = []
        for parameter_name, parameter_type in self.non_path_and_non_header_params.items():
            type_description = self.type_converter.get_openapi_type_of_parameter(parameter_type)
            param_description = self.method_properties.get_description_for_param(parameter_name)
            parameters.append(
                Parameter(name=parameter_name, in_=ParameterType.query, schema_=type_description, description=param_description)
            )
        return parameters

    # Request body

    def convert_request_body(self) -> RequestBody:
        properties = self._convert_function_params_to_openapi_request_body_properties()
        return self._build_json_request_body(properties)

    def _convert_function_params_to_openapi_request_body_properties(self) -> Dict[str, Schema]:
        properties = {}
        for parameter_name, parameter_type in self.non_path_and_non_header_params.items():
            type_description = self.type_converter.get_openapi_type_of_parameter(parameter_type)
            properties[parameter_name] = type_description
        return properties

    def _get_request_body_description(self) -> str:
        """
            OpenAPI supports only a single description field for the full request body. As such,
            this method return the description of the request body in CommonMark syntax to create
            an itemization of the different parameters in the request body.
        """
        result = ""
        for param_name in sorted(self.non_path_and_non_header_params.keys()):
            description = self.method_properties.get_description_for_param(param_name)
            if description is not None:
                result += f"* **{param_name}:** {description}\n"
            else:
                result += f"* **{param_name}:**\n"
        return result

    def _build_json_request_body(self, properties: Dict) -> RequestBody:
        request_body = RequestBody(
            required=True,
            content={"application/json": MediaType(schema=Schema(type="object", properties=properties))},
            description=self._get_request_body_description(),
        )
        return request_body


class OperationHandler:
    """
        Builds an OpenAPI Operation object from an Inmanta UrlMethod
    """

    def __init__(self, type_converter: OpenApiTypeConverter, arg_option_handler: ArgOptionHandler):
        self.type_converter = type_converter
        self.arg_option_handler = arg_option_handler

    def handle_method(self, url_method: UrlMethod, path: str) -> Operation:
        function_parameter_handler = FunctionParameterHandler(
            self.type_converter, self.arg_option_handler, path, url_method.properties
        )
        parameters = function_parameter_handler.get_parameters()
        responses = self._build_responses(url_method.properties)

        if url_method.get_operation() in ["POST", "PUT", "PATCH"]:
            extra_params = {"requestBody": function_parameter_handler.convert_request_body()}
        else:
            extra_params = {}

        tags = self._get_tags_of_operation(url_method)

        return Operation(
            responses=responses,
            operationId=url_method.method_name,
            parameters=(parameters if len(parameters) else None),
            summary=url_method.short_method_description,
            description=url_method.long_method_description,
            tags=tags,
            **extra_params,
        )

    def _get_tags_of_operation(self, url_method: UrlMethod) -> Optional[List[str]]:
        if url_method.endpoint is not None:
            if hasattr(url_method.endpoint, "_name"):
                return [url_method.endpoint._name]
            else:
                return [url_method.endpoint.__class__.__name__]
        else:
            return None

    def _build_responses(self, url_method_properties: MethodProperties) -> Dict[str, Response]:
        result: Dict[str, Response] = {}
        status_code_to_description_map = url_method_properties.get_description_foreach_http_status_code()
        for status_code, description in status_code_to_description_map.items():
            if status_code == 200:
                response_headers = self.arg_option_handler.extract_response_headers_from_arg_options(
                    url_method_properties.arg_options
                )
                return_value = self._build_return_value_wrapper(url_method_properties)
                result[str(status_code)] = Response(description=description, content=return_value, headers=response_headers)
            else:
                result[str(status_code)] = Response(description=description)

        return result

    def _build_return_value_wrapper(self, url_method_properties: MethodProperties) -> Optional[Dict[str, MediaType]]:
        return_type = inspect.signature(url_method_properties.function).return_annotation
        if return_type is not None and return_type != inspect.Signature.empty:
            return_properties: Optional[Dict[str, SchemaBase]] = None
            openapi_return_type = self.type_converter.get_openapi_type(return_type)
            if url_method_properties.envelope:
                return_properties = {url_method_properties.envelope_key: openapi_return_type}
            return {"application/json": MediaType(schema=Schema(type="object", properties=return_properties))}
        return None



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import datetime
import uuid
from typing import Any, Dict, List, NewType, Optional, Union

import pydantic

from inmanta import const
from inmanta.const import Change, ResourceState
from inmanta.types import ArgumentTypes, JsonType, SimpleTypes, StrictNonIntBool


class BaseModel(pydantic.BaseModel):
    """
        Base class for all data objects in Inmanta
    """

    class Config:
        # Populate models with the value property of enums, rather than the raw enum.
        # This is useful to serialise model.dict() later
        use_enum_values = True


class ExtensionStatus(BaseModel):
    """
        Status response for extensions loaded in the server
    """

    name: str
    version: str
    package: str


class SliceStatus(BaseModel):
    """
        Status response for slices loaded in the the server
    """

    name: str
    status: Dict[str, ArgumentTypes]


class FeatureStatus(BaseModel):
    """
        Status of the feature
    """

    slice: str
    name: str
    value: Optional[Any]


class StatusResponse(BaseModel):
    """
        Response for the status method call
    """

    product: str
    edition: str
    version: str
    license: Union[str, Dict[str, SimpleTypes]]
    extensions: List[ExtensionStatus]
    slices: List[SliceStatus]
    features: List[FeatureStatus]


class CompileRun(BaseModel):
    id: uuid.UUID
    remote_id: Optional[uuid.UUID]
    environment: uuid.UUID
    requested: Optional[datetime.datetime]
    started: Optional[datetime.datetime]

    do_export: bool
    force_update: bool
    metadata: JsonType
    environment_variables: Dict[str, str]


ResourceVersionIdStr = NewType("ResourceVersionIdStr", str)
"""
    The resource id with the version included.
"""


ResourceIdStr = NewType("ResourceIdStr", str)
"""
    The resource id without the version
"""


ResourceType = NewType("ResourceType", str)
"""
    The type of the resource
"""


class AttributeStateChange(BaseModel):
    """
        Changes in the attribute
    """

    current: Optional[Union[SimpleTypes, List[SimpleTypes], Dict[str, SimpleTypes]]] = None
    desired: Optional[Union[SimpleTypes, List[SimpleTypes], Dict[str, SimpleTypes]]] = None


class Event(BaseModel):
    """
        An event in the list of events generated by dependencies during deploy
    """

    status: ResourceState
    change: Change
    changes: Dict[str, AttributeStateChange]


EnvSettingType = Union[StrictNonIntBool, int, str, Dict[str, Union[str, int, StrictNonIntBool]]]


class Environment(BaseModel):
    """
        An inmanta environment.

        :note: repo_url and repo_branch will be moved to the settings.
    """

    id: uuid.UUID
    name: str
    project_id: uuid.UUID
    repo_url: str
    repo_branch: str
    settings: Dict[str, EnvSettingType]


class Project(BaseModel):
    """
        An inmanta environment.
    """

    id: uuid.UUID
    name: str
    environments: List[Environment]


class EnvironmentSetting(BaseModel):
    """ A class to define a new environment setting.

        :param name: The name of the setting.
        :param type: The type of the value. This type is mainly used for documentation purpose.
        :param default: An optional default value for this setting. When a default is set and the
                        is requested from the database, it will return the default value and also store
                        the default value in the database.
        :param doc: The documentation/help string for this setting
        :param recompile: Trigger a recompile of the model when a setting is updated?
        :param update_model: Update the configuration model (git pull on project and repos)
        :param agent_restart: Restart autostarted agents when this settings is updated.
        :param allowed_values: list of possible values (if type is enum)
    """

    name: str
    type: str
    default: EnvSettingType
    doc: str
    recompile: bool
    update_model: bool
    agent_restart: bool
    allowed_values: Optional[List[EnvSettingType]]


class EnvironmentSettingsReponse(BaseModel):

    settings: Dict[str, EnvSettingType]
    definition: Dict[str, EnvironmentSetting]


class ModelMetadata(BaseModel):
    """ Model metadata
    """

    inmanta_compile_state: const.Compilestate = const.Compilestate.success
    message: str
    type: str
    extra_data: Optional[JsonType]

    class Config:
        fields = {"inmanta_compile_state": {"alias": "inmanta:compile:state"}}


class ModelVersionInfo(BaseModel):
    """ Version information that can be associated with an orchestration model

        :param export_metadata: Metadata associated with this version
        :param model: A serialization of the complete orchestration model
    """

    export_metadata: ModelMetadata
    model: Optional[JsonType]


class Resource(BaseModel):
    environment: uuid.UUID
    model: int
    resource_id: ResourceVersionIdStr
    resource_type: ResourceType
    resource_version_id: ResourceVersionIdStr
    agent: str
    last_deploy: Optional[datetime.datetime]
    attributes: JsonType
    status: const.ResourceState



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import hashlib
import inspect
import os
import shutil
import sys

import pytest
from pytest import fixture

from inmanta import loader
from inmanta.module import Project


def test_code_manager():
    """ Verify the code manager
    """
    mgr = loader.CodeManager()
    mgr.register_code("test", test_code_manager)

    # get types
    types = mgr.get_types()
    name, type_list = next(types)
    assert name == "test"
    assert len(type_list) == 1

    source_info = type_list[0]

    def get_code():
        filename = inspect.getsourcefile(test_code_manager)
        with open(filename, "r", encoding="utf-8") as fd:
            return fd.read()

    content = get_code()
    assert source_info.content == content
    assert len(source_info.hash) > 0

    with pytest.raises(Exception):
        # only available in a valid project. Other test cases with full project validate this. This one checks
        # whether we give an error.
        source_info.requires()

    # get_file_hashes
    hashes = mgr.get_file_hashes()
    mgr_content = mgr.get_file_content(next(hashes))
    assert mgr_content == content

    with pytest.raises(KeyError):
        mgr.get_file_content("test")

    # register type without source
    with pytest.raises(loader.SourceNotFoundException):
        mgr.register_code("test2", str)


def test_empty_code_loader(tmp_path):
    """ Test loading an empty cache
    """
    cl = loader.CodeLoader(tmp_path)
    cl.load_modules()


def test_code_loader(tmp_path):
    """ Test loading a new module
    """
    cl = loader.CodeLoader(tmp_path)
    code = """
def test():
    return 10
    """

    sha1sum = hashlib.new("sha1")
    sha1sum.update(code.encode())
    hv = sha1sum.hexdigest()

    with pytest.raises(ImportError):
        import inmanta_unit_test  # NOQA

    cl.deploy_version(hv, "inmanta_unit_test", code)

    import inmanta_unit_test  # NOQA

    assert inmanta_unit_test.test() == 10

    # reload cached code
    cl.load_modules()


def test_code_loader_import_error(tmp_path, caplog):
    """ Test loading code with an import error
    """
    cl = loader.CodeLoader(tmp_path)
    code = """
import badimmport
def test():
    return 10
    """

    sha1sum = hashlib.new("sha1")
    sha1sum.update(code.encode())
    hv = sha1sum.hexdigest()

    with pytest.raises(ImportError):
        import inmanta_bad_unit_test  # NOQA

    cl.deploy_version(hv, "inmanta_bad_unit_test", code)

    assert "ModuleNotFoundError: No module named 'badimmport'" in caplog.text


@fixture(scope="function")
def module_path(tmpdir):
    module_finder = loader.PluginModuleFinder([str(tmpdir)])
    sys.meta_path.insert(0, module_finder)
    yield str(tmpdir)
    sys.meta_path.remove(module_finder)


def test_module_loader(module_path, tmpdir, capsys):
    """
        Verify that the loader.PluginModuleFinder and loader.PluginModuleLoader load modules correctly.
    """
    origin_mod_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "modules", "submodule")
    mod_dir = tmpdir.join(os.path.basename(origin_mod_dir))
    shutil.copytree(origin_mod_dir, mod_dir)

    capsys.readouterr()  # Clear buffers

    from inmanta_plugins.submodule import test

    assert test() == "test"
    (stdout, stderr) = capsys.readouterr()
    assert stdout.count("#loading inmanta_plugins.submodule#") == 1
    assert stdout.count("#loading inmanta_plugins.submodule.submod#") == 0
    assert stdout.count("#loading inmanta_plugins.submodule.pkg#") == 0
    assert stdout.count("#loading inmanta_plugins.submodule.pkg.submod2#") == 0

    from inmanta_plugins.submodule.submod import test_submod

    assert test_submod() == "test_submod"
    (stdout, stderr) = capsys.readouterr()
    assert stdout.count("#loading inmanta_plugins.submodule#") == 0
    assert stdout.count("#loading inmanta_plugins.submodule.submod#") == 1
    assert stdout.count("#loading inmanta_plugins.submodule.pkg#") == 0
    assert stdout.count("#loading inmanta_plugins.submodule.pkg.submod2#") == 0

    from inmanta_plugins.submodule.pkg import test_pkg

    assert test_pkg() == "test_pkg -- test_submod2"
    (stdout, stderr) = capsys.readouterr()
    assert stdout.count("#loading inmanta_plugins.submodule#") == 0
    assert stdout.count("#loading inmanta_plugins.submodule.submod#") == 0
    assert stdout.count("#loading inmanta_plugins.submodule.pkg#") == 1
    assert stdout.count("#loading inmanta_plugins.submodule.pkg.submod2#") == 1

    with pytest.raises(ImportError):
        from inmanta_plugins.tests import doesnotexist  # NOQA


def test_plugin_loading_on_project_load(tmpdir, capsys):
    """
        Load all plugins via the Project.load() method call and verify that no
        module is loaded twice when an import statement is used.
    """
    main_cf = tmpdir.join("main.cf")
    main_cf.write("import submodule")

    project_yml = tmpdir.join("project.yml")
    project_yml.write(
        """
name: test
modulepath: libs
downloadpath: libs
repo: https://github.com/inmanta/inmanta.git
install_mode: master
    """
    )

    tmpdir.mkdir("libs")
    origin_mod_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "modules", "submodule")
    mod_dir = tmpdir.join("libs", os.path.basename(origin_mod_dir))
    shutil.copytree(origin_mod_dir, mod_dir)

    project = Project(tmpdir, autostd=False)
    project.load()

    (stdout, stderr) = capsys.readouterr()
    assert stdout.count("#loading inmanta_plugins.submodule#") == 1
    assert stdout.count("#loading inmanta_plugins.submodule.submod#") == 1
    assert stdout.count("#loading inmanta_plugins.submodule.pkg#") == 1
    assert stdout.count("#loading inmanta_plugins.submodule.pkg.submod2#") == 1

    from inmanta_plugins.submodule import test

    assert test() == "test"
    (stdout, stderr) = capsys.readouterr()
    assert "#loading" not in stdout

    from inmanta_plugins.submodule.submod import test_submod

    assert test_submod() == "test_submod"
    (stdout, stderr) = capsys.readouterr()
    assert "#loading" not in stdout

    from inmanta_plugins.submodule.pkg import test_pkg

    assert test_pkg() == "test_pkg -- test_submod2"
    (stdout, stderr) = capsys.readouterr()
    assert "#loading" not in stdout



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import pytest

from inmanta import data
from inmanta.util import get_compiler_version


@pytest.mark.asyncio
async def test_environment_settings(client, server, environment_default):
    """
        Test environment settings
    """
    result = await client.list_settings(tid=environment_default)
    assert result.code == 200
    assert "settings" in result.result
    assert "metadata" in result.result
    assert "auto_deploy" in result.result["metadata"]
    assert len(result.result["settings"]) == 0

    # set invalid value
    result = await client.set_setting(tid=environment_default, id="auto_deploy", value="test")
    assert result.code == 500

    # set non existing setting
    result = await client.set_setting(tid=environment_default, id="auto_deploy_non", value=False)
    assert result.code == 404

    result = await client.set_setting(tid=environment_default, id="auto_deploy", value=False)
    assert result.code == 200

    result = await client.list_settings(tid=environment_default)
    assert result.code == 200
    assert len(result.result["settings"]) == 1

    result = await client.get_setting(tid=environment_default, id="auto_deploy")
    assert result.code == 200
    assert not result.result["value"]

    result = await client.get_setting(tid=environment_default, id="test2")
    assert result.code == 404

    result = await client.set_setting(tid=environment_default, id="auto_deploy", value=True)
    assert result.code == 200

    result = await client.get_setting(tid=environment_default, id="auto_deploy")
    assert result.code == 200
    assert result.result["value"]

    result = await client.delete_setting(tid=environment_default, id="test2")
    assert result.code == 404

    result = await client.delete_setting(tid=environment_default, id="auto_deploy")
    assert result.code == 200

    result = await client.list_settings(tid=environment_default)
    assert result.code == 200
    assert "settings" in result.result
    assert len(result.result["settings"]) == 1

    result = await client.set_setting(tid=environment_default, id=data.AUTOSTART_AGENT_DEPLOY_SPLAY_TIME, value=20)
    assert result.code == 200

    result = await client.set_setting(tid=environment_default, id=data.AUTOSTART_AGENT_DEPLOY_SPLAY_TIME, value="30")
    assert result.code == 200

    result = await client.get_setting(tid=environment_default, id=data.AUTOSTART_AGENT_DEPLOY_SPLAY_TIME)
    assert result.code == 200
    assert result.result["value"] == 30

    result = await client.delete_setting(tid=environment_default, id=data.AUTOSTART_AGENT_DEPLOY_SPLAY_TIME)
    assert result.code == 200

    agent_map = {"internal": "", "agent1": "", "agent2": "localhost", "agent3": "user@agent3"}
    result = await client.set_setting(tid=environment_default, id=data.AUTOSTART_AGENT_MAP, value=agent_map,)
    assert result.code == 200

    # Internal agent is missing
    result = await client.set_setting(tid=environment_default, id=data.AUTOSTART_AGENT_MAP, value={"agent1": ""})
    assert result.code == 500
    assert "The internal agent must be present in the autostart_agent_map" in result.result["message"]
    # Assert agent_map didn't change
    result = await client.get_setting(tid=environment_default, id=data.AUTOSTART_AGENT_MAP)
    assert result.code == 200
    assert result.result["value"] == agent_map

    result = await client.set_setting(tid=environment_default, id=data.AUTOSTART_AGENT_MAP, value="")
    assert result.code == 500
    assert "Agent map should be a dict" in result.result["message"]
    # Assert agent_map didn't change
    result = await client.get_setting(tid=environment_default, id=data.AUTOSTART_AGENT_MAP)
    assert result.code == 200
    assert result.result["value"] == agent_map


@pytest.mark.asyncio
async def test_environment_settings_v2(client_v2, server, environment_default):
    """
        Test environment settings
    """
    response = await client_v2.environment_settings_list(tid=environment_default)
    assert response.code == 200
    assert "settings" in response.result["data"]
    assert "definition" in response.result["data"]
    assert "auto_deploy" in response.result["data"]["definition"]
    assert len(response.result["data"]["settings"]) == 0

    response = await client_v2.environment_settings_set(tid=environment_default, id="auto_deploy", value=False)
    assert response.code == 200

    response = await client_v2.environment_settings_set(tid=environment_default, id="auto_deploy2", value=False)
    assert response.code == 404

    response = await client_v2.environment_settings_set(tid=environment_default, id="auto_deploy", value="error")
    assert response.code == 500

    response = await client_v2.environment_setting_delete(tid=environment_default, id="auto_deploy")
    assert response.code == 200

    response = await client_v2.environment_setting_delete(tid=environment_default, id="auto_deploy2")
    assert response.code == 404


@pytest.mark.asyncio
async def test_delete_protected_environment(server, client):
    result = await client.create_project("env-test")
    assert result.code == 200
    project_id = result.result["project"]["id"]

    async def create_environment() -> str:
        result = await client.create_environment(project_id=project_id, name="dev")
        assert result.code == 200
        return result.result["environment"]["id"]

    async def assert_env_deletion(env_id: str, deletion_succeeds: bool) -> None:
        # Execute delete operation
        result = await client.environment_delete(env_id)
        assert result.code == 200 if deletion_succeeds else 403
        # Assert result
        result = await client.environment_get(env_id)
        assert result.code == 404 if deletion_succeeds else 200

    # Test default setting
    env_id = await create_environment()
    await assert_env_deletion(env_id, deletion_succeeds=True)

    # Test environment is protected
    env_id = await create_environment()
    result = await client.environment_settings_set(env_id, data.PROTECTED_ENVIRONMENT, True)
    assert result.code == 200
    await assert_env_deletion(env_id, deletion_succeeds=False)

    # Test environment is unprotected
    result = await client.environment_settings_set(env_id, data.PROTECTED_ENVIRONMENT, False)
    assert result.code == 200
    await assert_env_deletion(env_id, deletion_succeeds=True)


@pytest.mark.asyncio
async def test_clear_protected_environment(server, client):
    result = await client.create_project("env-test")
    assert result.code == 200
    project_id = result.result["project"]["id"]

    result = await client.create_environment(project_id=project_id, name="dev")
    assert result.code == 200
    env_id = result.result["environment"]["id"]

    async def push_version_to_environment() -> None:
        res = await client.reserve_version(env_id)
        assert res.code == 200
        version = res.result["data"]

        result = await client.put_version(
            tid=env_id, version=version, resources=[], unknowns=[], compiler_version=get_compiler_version(),
        )
        assert result.code == 200

    async def assert_clear_env(env_id: str, clear_succeeds: bool) -> None:
        result = await client.list_versions(env_id)
        assert result.code == 200
        assert len(result.result["versions"]) != 0
        # Execute clear operation
        result = await client.environment_clear(env_id)
        assert result.code == 200 if clear_succeeds else 403
        # Assert result
        result = await client.list_versions(env_id)
        assert result.code == 200
        assert (len(result.result["versions"]) == 0) == clear_succeeds

    # Test default settings
    await push_version_to_environment()
    await assert_clear_env(env_id, clear_succeeds=True)

    # Test environment is protected
    await push_version_to_environment()
    result = await client.environment_settings_set(env_id, data.PROTECTED_ENVIRONMENT, True)
    assert result.code == 200
    await assert_clear_env(env_id, clear_succeeds=False)

    # Test environment is unprotected
    result = await client.environment_settings_set(env_id, data.PROTECTED_ENVIRONMENT, False)
    assert result.code == 200
    await assert_clear_env(env_id, clear_succeeds=True)



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import asyncio
import logging
import ssl
import uuid
from asyncio import CancelledError
from collections import defaultdict
from json import JSONDecodeError
from typing import Dict, List, MutableMapping, Optional, Union

import tornado
from pyformance import timer
from tornado import httpserver, iostream, routing, web

import inmanta.protocol.endpoints
from inmanta import config as inmanta_config
from inmanta import const
from inmanta.protocol import common, exceptions
from inmanta.protocol.rest import RESTBase
from inmanta.server import config as server_config
from inmanta.server.config import server_access_control_allow_origin, server_enable_auth
from inmanta.types import ReturnTypes

LOGGER: logging.Logger = logging.getLogger(__name__)


class RESTHandler(tornado.web.RequestHandler):
    """
        A generic class use by the transport
    """

    def initialize(self, transport: "RESTServer", config: Dict[str, common.UrlMethod]) -> None:
        self._transport: "RESTServer" = transport
        self._config = config

    def _get_config(self, http_method: str) -> common.UrlMethod:
        if http_method.upper() not in self._config:
            allowed = ", ".join(self._config.keys())
            self.set_header("Allow", allowed)
            raise exceptions.BaseHttpException(
                405, "%s is not supported for this url. Supported methods: %s" % (http_method, allowed)
            )

        return self._config[http_method]

    def get_auth_token(self, headers: MutableMapping[str, str]) -> Optional[MutableMapping[str, str]]:
        """
            Get the auth token provided by the caller. The token is provided as a bearer token.
        """
        if "Authorization" not in headers:
            return None

        parts = headers["Authorization"].split(" ")
        if len(parts) == 0 or parts[0].lower() != "bearer" or len(parts) > 2 or len(parts) == 1:
            LOGGER.warning(
                "Invalid authentication header, Inmanta expects a bearer token. (%s was provided)", headers["Authorization"]
            )
            return None

        return common.decode_token(parts[1])

    def prepare(self) -> None:
        # Setting "Access-Control-Allow-Origin": null can be exploited.
        # better not set it all instead.
        # See: https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Origin
        server_origin = server_access_control_allow_origin.get()
        if server_origin is not None:
            self.set_header("Access-Control-Allow-Origin", server_origin)

    def respond(self, body: ReturnTypes, headers: MutableMapping[str, str], status: int) -> None:
        if common.CONTENT_TYPE not in headers:
            headers[common.CONTENT_TYPE] = common.JSON_CONTENT

        if body is not None:
            encoded_body = self._encode_body(body, headers[common.CONTENT_TYPE])
            self.write(encoded_body)

        for header, value in headers.items():
            self.set_header(header, value)

        self.set_status(status)

    def _encode_body(self, body: ReturnTypes, content_type: str) -> Union[ReturnTypes, bytes]:
        if content_type == common.JSON_CONTENT:
            return common.json_encode(body)
        if content_type == common.HTML_CONTENT:
            return body.encode(common.HTML_ENCODING)
        if content_type == common.HTML_CONTENT_WITH_UTF8_CHARSET:
            return body.encode(common.UTF8_ENCODING)
        return body

    async def _call(self, kwargs: Dict[str, str], http_method: str, call_config: common.UrlMethod) -> None:
        """
            An rpc like call
        """
        if call_config is None:
            raise exceptions.NotFound("This method does not exist")

        if not self._transport.running:
            return

        with timer("rpc." + call_config.method_name).time():
            self._transport.start_request()
            try:
                message = self._transport._decode(self.request.body)
                LOGGER.debug(f"HTTP version of request: {self.request.version}")
                if message is None:
                    message = {}

                for key, value in self.request.query_arguments.items():
                    if len(value) == 1:
                        message[key] = value[0].decode("latin-1")
                    else:
                        message[key] = [v.decode("latin-1") for v in value]

                request_headers = self.request.headers
                auth_token = self.get_auth_token(request_headers)

                auth_enabled = server_enable_auth.get()
                if not auth_enabled or auth_token is not None:
                    result = await self._transport._execute_call(
                        kwargs, http_method, call_config, message, request_headers, auth_token
                    )
                    self.respond(result.body, result.headers, result.status_code)
                else:
                    raise exceptions.UnauthorizedException("Access to this resource is unauthorized.")

            except JSONDecodeError as e:
                error_message = f"The request body couldn't be decoded as a JSON: {e}"
                LOGGER.info(error_message, exc_info=True)
                self.respond({"message": error_message}, {}, 400)

            except ValueError:
                LOGGER.exception("An exception occured")
                self.respond({"message": "Unable to decode request body"}, {}, 400)

            except exceptions.BaseHttpException as e:
                self.respond(e.to_body(), {}, e.to_status())

            except CancelledError:
                self.respond({"message": "Request is cancelled on the server"}, {}, 500)

            finally:
                try:
                    await self.finish()
                except iostream.StreamClosedError:
                    # The connection has been closed already.
                    pass
                except Exception:
                    LOGGER.exception("An exception occurred responding to %s", self.request.remote_ip)
                self._transport.end_request()

    async def head(self, *args: str, **kwargs: str) -> None:
        if args:
            raise Exception("Only named groups are support in url patterns")
        await self._transport.add_background_task(
            self._call(http_method="HEAD", call_config=self._get_config("HEAD"), kwargs=kwargs), cancel_on_stop=False
        )

    async def get(self, *args: str, **kwargs: str) -> None:
        if args:
            raise Exception("Only named groups are support in url patterns")
        await self._transport.add_background_task(
            self._call(http_method="GET", call_config=self._get_config("GET"), kwargs=kwargs), cancel_on_stop=False
        )

    async def post(self, *args: str, **kwargs: str) -> None:
        if args:
            raise Exception("Only named groups are support in url patterns")
        await self._transport.add_background_task(
            self._call(http_method="POST", call_config=self._get_config("POST"), kwargs=kwargs), cancel_on_stop=False
        )

    async def delete(self, *args: str, **kwargs: str) -> None:
        if args:
            raise Exception("Only named groups are support in url patterns")
        await self._transport.add_background_task(
            self._call(http_method="DELETE", call_config=self._get_config("DELETE"), kwargs=kwargs), cancel_on_stop=False
        )

    async def patch(self, *args: str, **kwargs: str) -> None:
        if args:
            raise Exception("Only named groups are support in url patterns")
        await self._transport.add_background_task(
            self._call(http_method="PATCH", call_config=self._get_config("PATCH"), kwargs=kwargs), cancel_on_stop=False
        )

    async def put(self, *args: str, **kwargs: str) -> None:
        if args:
            raise Exception("Only named groups are support in url patterns")
        await self._transport.add_background_task(
            self._call(http_method="PUT", call_config=self._get_config("PUT"), kwargs=kwargs), cancel_on_stop=False
        )

    def options(self, *args: str, **kwargs: str) -> None:
        if args:
            raise Exception("Only named groups are support in url patterns")

        allow_headers = "Origin, Accept, Content-Type, X-Requested-With, X-CSRF-Token, %s" % const.INMANTA_MT_HEADER
        if len(self._transport.headers):
            allow_headers += ", " + ", ".join(self._transport.headers)
        if server_enable_auth.get():
            allow_headers += ", Authorization"

        self.set_header("Access-Control-Allow-Methods", "HEAD, GET, POST, PUT, OPTIONS, DELETE, PATCH")
        self.set_header("Access-Control-Allow-Headers", allow_headers)

        self.set_status(200)


class StaticContentHandler(tornado.web.RequestHandler):
    def initialize(self, transport: "RESTServer", content: str, content_type: str) -> None:
        self._transport = transport
        self._content = content
        self._content_type = content_type

    def get(self, *args: List[str], **kwargs: Dict[str, str]) -> None:
        self.set_header("Content-Type", self._content_type)
        self.write(self._content)
        self.set_status(200)


class RESTServer(RESTBase):
    """
        A tornado based rest server
    """

    _http_server: Optional[httpserver.HTTPServer]

    def __init__(self, session_manager: common.SessionManagerInterface, id: str) -> None:
        super().__init__()

        self._id = id
        self.headers: Dict[str, str] = {}
        self.session_manager = session_manager
        # number of ongoing requests
        self.inflight_counter = 0
        # event indicating no more in flight requests
        self.idle_event = asyncio.Event()
        self.idle_event.set()
        self.running = False
        self._http_server = None

    def start_request(self):
        self.idle_event.clear()
        self.inflight_counter += 1

    def end_request(self):
        self.inflight_counter -= 1
        if self.inflight_counter == 0:
            self.idle_event.set()

    def validate_sid(self, sid: uuid.UUID) -> bool:
        return self.session_manager.validate_sid(sid)

    def get_global_url_map(
        self, targets: List[inmanta.protocol.endpoints.CallTarget]
    ) -> Dict[str, Dict[str, common.UrlMethod]]:
        global_url_map: Dict[str, Dict[str, common.UrlMethod]] = defaultdict(dict)
        for slice in targets:
            url_map = slice.get_op_mapping()
            for url, configs in url_map.items():
                handler_config = global_url_map[url]
                for op, cfg in configs.items():
                    handler_config[op] = cfg
        return global_url_map

    async def start(
        self, targets: List[inmanta.protocol.endpoints.CallTarget], additional_rules: List[routing.Rule] = []
    ) -> None:
        """
            Start the server on the current ioloop
        """
        global_url_map: Dict[str, Dict[str, common.UrlMethod]] = self.get_global_url_map(targets)

        rules: List[routing.Rule] = []
        rules.extend(additional_rules)

        for url, handler_config in global_url_map.items():
            rules.append(routing.Rule(routing.PathMatches(url), RESTHandler, {"transport": self, "config": handler_config}))
            LOGGER.debug("Registering handler(s) for url %s and methods %s", url, ", ".join(handler_config.keys()))

        application = web.Application(rules, compress_response=True)

        crt = inmanta_config.Config.get("server", "ssl_cert_file", None)
        key = inmanta_config.Config.get("server", "ssl_key_file", None)

        if crt is not None and key is not None:
            ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_ctx.load_cert_chain(crt, key)

            self._http_server = httpserver.HTTPServer(application, decompress_request=True, ssl_options=ssl_ctx)
            LOGGER.debug("Created REST transport with SSL")
        else:
            self._http_server = httpserver.HTTPServer(application, decompress_request=True)

        bind_port = server_config.get_bind_port()
        bind_addresses = server_config.server_bind_address.get()

        for bind_addr in bind_addresses:
            self._http_server.listen(bind_port, bind_addr)
            LOGGER.info(f"Server listening on {bind_addr}:{bind_port}")
        self.running = True

        LOGGER.debug("Start REST transport")

    async def stop(self) -> None:
        """
            Stop the current server
        """
        self.running = False
        LOGGER.debug("Stopping Server Rest Endpoint")
        if self._http_server is not None:
            self._http_server.stop()

    async def join(self) -> None:
        await self.idle_event.wait()
        if self._http_server is not None:
            await self._http_server.close_all_connections()



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import re
import warnings
from typing import List

import pytest

from inmanta.ast import LocatableString, Namespace
from inmanta.ast.blocks import BasicBlock
from inmanta.ast.constraint.expression import And, Equals, GreaterThan, In, IsDefined, Not, Or, Regex
from inmanta.ast.statements import ExpressionStatement, Literal, ReferenceStatement, define
from inmanta.ast.statements.assign import (
    Assign,
    CreateDict,
    CreateList,
    IndexLookup,
    MapLookup,
    SetAttribute,
    ShortIndexLookup,
    StringFormat,
)
from inmanta.ast.statements.call import FunctionCall
from inmanta.ast.statements.define import (
    DefineEntity,
    DefineImplement,
    DefineIndex,
    DefineTypeConstraint,
    DefineTypeDefault,
    TypeDeclaration,
)
from inmanta.ast.statements.generator import Constructor, If
from inmanta.ast.variables import AttributeReference, Reference
from inmanta.execute.util import NoneValue
from inmanta.parser import ParserException, SyntaxDeprecationWarning
from inmanta.parser.plyInmantaParser import parse


def parse_code(model_code: str):
    root_ns = Namespace("__root__")
    main_ns = Namespace("__config__")
    main_ns.parent = root_ns
    statements = parse(main_ns, "test", model_code)

    return statements


def test_define_empty():
    parse_code("""""")


def test_define_entity():
    """Test the definition of entities
    """
    statements = parse_code(
        """
entity Test:
end
entity Other:
string hello
end
entity Other:
 \"\"\"XX
 \"\"\"
end
"""
    )

    assert len(statements) == 3

    stmt = statements[0]
    assert isinstance(stmt, define.DefineEntity)
    assert stmt.name == "Test"
    assert [str(p) for p in stmt.parents] == ["std::Entity"]
    assert len(stmt.attributes) == 0
    assert stmt.comment is None
    assert stmt.type.comment is None


def test_undefine_default():
    statements = parse_code(
        """
entity Test extends Foo:
 string hello = undef
 string[] dinges = undef
end"""
    )
    assert len(statements) == 1

    stmt = statements[0]
    assert isinstance(stmt, define.DefineEntity)
    assert stmt.name == "Test"
    assert [str(p) for p in stmt.parents] == ["Foo"]
    assert len(stmt.attributes) == 2
    assert stmt.comment is None
    assert stmt.type.comment is None

    for ad in stmt.attributes:
        assert isinstance(ad.type, TypeDeclaration)
        assert isinstance(ad.type.basetype, LocatableString)
        assert isinstance(ad.name, LocatableString)
        assert ad.default is None
        assert ad.remove_default

    assert str(stmt.attributes[0].name) == "hello"
    assert str(stmt.attributes[1].name) == "dinges"


def test_extend_entity():
    """Test extending entities
    """
    statements = parse_code(
        """
entity Test extends Foo:
end
"""
    )

    assert len(statements) == 1

    stmt = statements[0]
    assert [str(p) for p in stmt.parents] == ["Foo"]


def test_complex_entity():
    """Test definition of a complex entity
    """
    documentation = "This entity has documentation"
    statements = parse_code(
        """
entity Test extends Foo, foo::sub::Bar:
    \"\"\" %s
    \"\"\"
    string hello
    bool bar = true
    number? ten=5
end
"""
        % documentation
    )

    assert len(statements) == 1

    stmt = statements[0]
    assert len(stmt.parents) == 2
    assert [str(p) for p in stmt.parents] == ["Foo", "foo::sub::Bar"]
    assert str(stmt.comment).strip() == documentation
    assert str(stmt.type.comment).strip() == documentation
    assert len(stmt.attributes) == 3

    for ad in stmt.attributes:
        assert isinstance(ad.type, TypeDeclaration)
        assert isinstance(ad.type.basetype, LocatableString)
        assert isinstance(ad.name, LocatableString)

    assert str(stmt.attributes[0].name) == "hello"
    assert str(stmt.attributes[1].name) == "bar"
    assert str(stmt.attributes[2].name) == "ten"

    assert stmt.attributes[1].default.execute(None, None, None)

    assert stmt.attributes[2].default.execute(None, None, None) == 5


def test_relation():
    """Test definition of relations
    """
    statements = parse_code(
        """
Test tests [0:] -- [5:10] Foo bars
"""
    )

    assert len(statements) == 1
    rel = statements[0]

    assert len(rel.left) == 3
    assert len(rel.right) == 3

    assert str(rel.left[0]) == "Test"
    assert str(rel.right[0]) == "Foo"

    assert str(rel.left[1]) == "tests"
    assert str(rel.right[1]) == "bars"

    assert rel.left[2] == (0, None)
    assert rel.right[2] == (5, 10)


def test_relation_2():
    """Test definition of relations
    """
    statements = parse_code(
        """
Test tests [3] -- [:10] Foo bars
"""
    )

    assert len(statements) == 1
    rel = statements[0]

    assert len(rel.left) == 3
    assert len(rel.right) == 3

    assert str(rel.left[0]) == "Test"
    assert str(rel.right[0]) == "Foo"

    assert str(rel.left[1]) == "tests"
    assert str(rel.right[1]) == "bars"

    assert rel.left[2] == (3, 3)
    assert rel.right[2] == (None, 10)


def test_new_relation():
    """Test definition of relations
    """
    statements = parse_code(
        """
Test.bar [1] -- Foo.tests [5:10]
"""
    )

    assert len(statements) == 1, "Should return four statements"
    rel = statements[0]

    assert len(rel.left) == 3
    assert len(rel.right) == 3

    assert str(rel.left[0]) == "Test"
    assert str(rel.right[0]) == "Foo"

    assert str(rel.left[1]) == "tests"
    assert str(rel.right[1]) == "bar"

    assert rel.left[2] == (5, 10)
    assert rel.right[2] == (1, 1)


def test_new_relation_with_annotations():
    """Test definition of relations
    """
    statements = parse_code(
        """
Test.bar [1] foo,bar Foo.tests [5:10]
"""
    )

    assert len(statements) == 1, "Should return four statements"
    rel = statements[0]

    assert len(rel.left) == 3
    assert len(rel.right) == 3

    assert str(rel.left[0]) == "Test"
    assert str(rel.right[0]) == "Foo"

    assert str(rel.left[1]) == "tests"
    assert str(rel.right[1]) == "bar"

    assert rel.left[2] == (5, 10)
    assert rel.right[2] == (1, 1)
    assert len(rel.annotations) == 2
    assert rel.annotation_expression[0][1].name == "foo"
    assert rel.annotation_expression[1][1].name == "bar"


def test_new_relation_unidir():
    """Test definition of relations
    """
    statements = parse_code(
        """
Test.bar [1] -- Foo
"""
    )

    assert len(statements) == 1, "Should return four statements"
    rel = statements[0]

    assert len(rel.left) == 3
    assert len(rel.right) == 3

    assert str(rel.left[0]) == "Test"
    assert str(rel.right[0]) == "Foo"

    assert (rel.left[1]) is None
    assert str(rel.right[1]) == "bar"

    assert rel.left[2] is None
    assert rel.right[2] == (1, 1)


def test_new_relation_with_annotations_unidir():
    """Test definition of relations
    """
    statements = parse_code(
        """
Test.bar [1] foo,bar Foo
"""
    )

    assert len(statements) == 1, "Should return four statements"
    rel = statements[0]

    assert len(rel.left) == 3
    assert len(rel.right) == 3

    assert str(rel.left[0]) == "Test"
    assert str(rel.right[0]) == "Foo"

    assert rel.left[1] is None
    assert str(rel.right[1]) == "bar"

    assert rel.left[2] is None
    assert rel.right[2] == (1, 1)
    assert len(rel.annotations) == 2
    assert rel.annotation_expression[0][1].name == "foo"
    assert rel.annotation_expression[1][1].name == "bar"


def test_implementation():
    """Test the definition of implementations
    """
    statements = parse_code(
        """
implementation test for Test:
end
"""
    )

    assert len(statements) == 1
    assert len(statements[0].block.get_stmts()) == 0
    assert statements[0].name == "test"
    assert isinstance(statements[0].entity, LocatableString)

    statements = parse_code(
        """
implementation test for Test:
    std::File(attr="a")
    var = hello::func("world")
end
"""
    )

    assert len(statements) == 1
    assert len(statements[0].block.get_stmts()) == 2


def test_implementation_with_for():
    """Test the propagation of type requires when using a for
    """
    statements = parse_code(
        """
implementation test for Test:
    \"\"\" test \"\"\"
    for v in data:
        std::template("template")
    end
end
"""
    )

    assert len(statements) == 1
    assert len(statements[0].block.get_stmts()) == 1


def test_implements():
    """Test implements with no selector
    """
    statements = parse_code(
        """
implement Test using test
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineImplement)
    assert str(stmt.entity) == "Test"
    assert [str(i) for i in stmt.implementations] == ["test"]
    assert str(stmt.select) == "true"


def test_implements_2():
    """Test implements with selector
    """
    statements = parse_code(
        """
implement Test using test, blah when (self > 5)
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineImplement)
    assert str(stmt.entity) == "Test"
    assert [str(p) for p in stmt.implementations] == ["test", "blah"]
    assert isinstance(stmt.select, GreaterThan)
    assert stmt.select.children[0].name == "self"
    assert stmt.select.children[1].value == 5


def test_implements_parent():
    statements = parse_code(
        """
implement Test using parents  \""" testc \"""
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineImplement)
    assert str(stmt.entity) == "Test"
    assert stmt.inherit is True


@pytest.mark.parametrize(
    "implementations",
    [["parents", "std::none"], ["std::none", "parents"], ["i1", "parents", "i2"], ["std::none"], ["i1", "i2"]],
)
def test_implements_parent_in_list(implementations: List[str]):
    statements = parse_code(
        """
implement Test using %s
        """
        % ", ".join(implementations)
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineImplement)
    assert str(stmt.entity) == "Test"
    assert stmt.inherit is ("parents" in implementations)
    assert [str(i) for i in stmt.implementations] == [i for i in implementations if i != "parents"]


def test_implements_selector():
    """Test implements with selector
    """
    statements = parse_code(
        """
implement Test using test when not (fg(self) and false)
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineImplement)
    assert str(stmt.entity) == "Test"
    assert [str(i) for i in stmt.implementations] == ["test"]
    assert isinstance(stmt.select, Not)
    assert isinstance(stmt.select.children[0], And)
    assert isinstance(stmt.select.children[0].children[0], FunctionCall)
    assert isinstance(stmt.select.children[0].children[1], Literal)


def test_regex():
    statements = parse_code(
        """
a = /[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}/
"""
    )

    assert len(statements) == 1
    stmt = statements[0].value
    assert isinstance(stmt, Regex)
    assert stmt.children[1].value == re.compile(r"[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}")


def test_regex_backslash():
    statements = parse_code(
        r"""
a = /\\/
"""
    )

    assert len(statements) == 1
    stmt = statements[0].value
    assert isinstance(stmt, Regex)
    assert stmt.children[1].value == re.compile(r"\\")


def test_regex_escape():
    statements = parse_code(
        r"""
a = /\/1/
"""
    )

    assert len(statements) == 1
    stmt = statements[0].value
    assert isinstance(stmt, Regex)
    assert stmt.children[1].value == re.compile(r"\/1")


def test_regex_twice():
    statements = parse_code(
        r"""
a = /\/1/
b = "v"
c = /\/1/
"""
    )

    assert len(statements) == 3
    stmt = statements[0].value
    assert isinstance(stmt, Regex)
    assert stmt.children[1].value == re.compile(r"\/1")


def test_1584_regex_error():
    with pytest.raises(ParserException) as pytest_e:
        parse_code(
            """
a = /)/
            """
        )

    exception: ParserException = pytest_e.value
    assert exception.location.file == "test"
    assert exception.location.lnr == 2
    assert exception.location.start_char == 5
    assert exception.location.end_lnr == 2
    assert exception.location.end_char == 8
    assert exception.value == "/)/"
    assert exception.msg == "Syntax error: Regex error in /)/: 'unbalanced parenthesis at position 0'"


def test_typedef():
    statements = parse_code(
        """
typedef uuid as string matching /[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}/
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineTypeConstraint)
    assert str(stmt.name) == "uuid"
    assert str(stmt.basetype) == "string"
    assert isinstance(stmt.get_expression(), Regex)
    assert stmt.get_expression().children[1].value == re.compile(
        r"[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}"
    )


def test_typedef_in():
    statements = parse_code(
        """
typedef abc as string matching self in ["a","b","c"]
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineTypeConstraint)
    assert str(stmt.name) == "abc"
    assert str(stmt.basetype) == "string"
    assert isinstance(stmt.get_expression(), In)
    assert [x.value for x in stmt.get_expression().children[1].items] == ["a", "b", "c"]


def test_typedef_plugin_call():
    """
    If this test fails, the collection of validation types and
    validation_parameters will fail in the LSM module.
    plugin_call() must expand to (plugin_call() == true)
    """
    statements = parse_code(
        """
typedef abc as string matching std::is_base64_encoded(self)
"""
    )
    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineTypeConstraint)
    assert str(stmt.name) == "abc"
    assert str(stmt.basetype) == "string"
    assert isinstance(stmt.get_expression(), Equals)
    left_side_equals = stmt.get_expression()._arguments[0]
    right_side_equals = stmt.get_expression()._arguments[1]
    assert isinstance(left_side_equals, FunctionCall)
    assert isinstance(right_side_equals, Literal)
    assert isinstance(right_side_equals.value, bool) and right_side_equals.value


def test_typedef2():
    statements = parse_code(
        """
typedef ConfigFile as File(mode = 644, owner = "root", group = "root")
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineTypeDefault)
    assert stmt.name == "ConfigFile"
    assert isinstance(stmt.ctor, Constructor)


def test_index():
    statements = parse_code(
        """
index File(host, path)
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineIndex)
    assert str(stmt.type) == "File"
    assert stmt.attributes == ["host", "path"]


def test_ctr():
    statements = parse_code(
        """
File(host = 5, path = "Jos")
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Constructor)
    assert str(stmt.class_type) == "File"
    assert {k: v.value for k, v in stmt.attributes.items()} == {"host": 5, "path": "Jos"}


def test_ctr_dict():
    statements = parse_code(
        """
dct = { "host": "myhost", "path": "/dir/file" }
File(**dct)
"""
    )

    assert len(statements) == 2
    stmt = statements[1]
    assert isinstance(stmt, Constructor)
    assert str(stmt.class_type) == "File"
    assert stmt.attributes == {}
    assert len(stmt.wrapped_kwargs) == 1


def test_ctr_dict_multi_param():
    statements = parse_code(
        """
dct = { "host": "myhost" }
File(**dct, path = "/dir/file")
"""
    )

    assert len(statements) == 2
    stmt = statements[1]
    assert isinstance(stmt, Constructor)
    assert str(stmt.class_type) == "File"
    assert {k: v.value for k, v in stmt.attributes.items()} == {"path": "/dir/file"}
    assert len(stmt.wrapped_kwargs) == 1


def test_ctr_dict_multi_param3():
    statements = parse_code(
        """
dct = { "host": "myhost" }
File(path = "/dir/file", **dct)
"""
    )

    assert len(statements) == 2
    stmt = statements[1]
    assert isinstance(stmt, Constructor)
    assert str(stmt.class_type) == "File"
    assert {k: v.value for k, v in stmt.attributes.items()} == {"path": "/dir/file"}
    assert len(stmt.wrapped_kwargs) == 1


def test_indexlookup():
    statements = parse_code(
        """
a=File[host = 5, path = "Jos"]
"""
    )

    assert len(statements) == 1
    stmt = statements[0].value
    assert isinstance(stmt, IndexLookup)
    assert str(stmt.index_type) == "File"
    assert {k: v.value for k, v in stmt.query} == {"host": 5, "path": "Jos"}


def test_indexlookup_kwargs():
    statements = parse_code(
        """
dct = {"path": "/dir/file"}
a=File[host = "myhost", **dct]
"""
    )

    assert len(statements) == 2
    stmt = statements[1].value
    assert isinstance(stmt, IndexLookup)
    assert str(stmt.index_type) == "File"
    assert {k: v.value for k, v in stmt.query} == {"host": "myhost"}
    assert len(stmt.wrapped_query) == 1


def test_short_index_lookup():
    statements = parse_code(
        """
a = vm.files[path="/etc/motd"]
"""
    )

    assert len(statements) == 1
    stmt = statements[0].value
    assert isinstance(stmt, ShortIndexLookup)
    assert isinstance(stmt.rootobject, Reference)
    assert stmt.rootobject.name == "vm"
    assert stmt.relation == "files"
    assert {k: v.value for k, v in stmt.querypart} == {"path": "/etc/motd"}


def test_short_index_lookup_kwargs():
    statements = parse_code(
        """
dct = {"path": "/etc/motd"}
a = vm.files[**dct]
"""
    )

    assert len(statements) == 2
    stmt = statements[1].value
    assert isinstance(stmt, ShortIndexLookup)
    assert isinstance(stmt.rootobject, Reference)
    assert stmt.rootobject.name == "vm"
    assert stmt.relation == "files"
    assert stmt.querypart == []
    assert len(stmt.wrapped_querypart) == 1


def test_ctr_2():
    statements = parse_code(
        """
File( )
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Constructor)
    assert str(stmt.class_type) == "File"
    assert {k: v.value for k, v in stmt.attributes.items()} == {}


def test_function():
    statements = parse_code(
        """
file( )
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, FunctionCall)
    assert str(stmt.name) == "file"


def test_function_2():
    statements = parse_code(
        """
file(b)
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, FunctionCall)
    assert str(stmt.name) == "file"


def test_function_3():
    statements = parse_code(
        """
file(b,)
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, FunctionCall)
    assert str(stmt.name) == "file"


def test_list_def():
    statements = parse_code(
        """
a=["a]","b"]
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, CreateList)
    assert [x.value for x in stmt.value.items] == ["a]", "b"]


def test_list_def_trailing_comma():
    statements = parse_code(
        """
a=["a]","b",]
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, CreateList)
    assert [x.value for x in stmt.value.items] == ["a]", "b"]


def test_map_def():
    statements = parse_code(
        """
a={ "a":"b", "b":1}
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, CreateDict)
    assert [(x[0], x[1].value) for x in stmt.value.items] == [("a", "b"), ("b", 1)]


def test_map_def_var():
    statements = parse_code("""a={ "b":b}""")
    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, CreateDict)
    assert isinstance(stmt.value.items[0][1], Reference)


def test_map_def_list():
    statements = parse_code("""a={ "a":["a"]}""")
    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, CreateDict)
    assert isinstance(stmt.value.items[0][1], CreateList)


def test_map_def_map():
    statements = parse_code("""a={ "a":{"a":"C"}}""")
    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, CreateDict)
    assert isinstance(stmt.value.items[0][1], CreateDict)


def test_booleans():
    statements = parse_code(
        """
a=true b=false
"""
    )

    assert len(statements) == 2
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert stmt.value.value
    assert not statements[1].value.value


def test_none():
    statements = parse_code(
        """
a=null
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value.value, NoneValue)


def test_numbers():
    statements = parse_code(
        """
a=1
b=2.0
c=-5
d=-0.256
"""
    )

    assert len(statements) == 4
    values = [1, 2.0, -5, -0.256]
    for i in range(4):
        stmt = statements[i]
        assert isinstance(stmt, Assign)
        assert stmt.value.value == values[i]


def test_string():
    statements = parse_code(
        """
a="jos"
"""
    )
    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, Literal)
    assert stmt.value.value == "jos"


def test_string_2():
    statements = parse_code(
        """
a='jos'
"""
    )
    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, Literal)
    assert stmt.value.value == "jos"


def test_string_backslash():
    statements = parse_code(
        """
a="\\\\"
"""
    )
    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, Literal)
    assert stmt.value.value == "\\"


def test_string_backslash_2():
    statements = parse_code(
        """
a='\\\\'
"""
    )
    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, Literal)
    assert stmt.value.value == "\\"


def test_empty():
    statements = parse_code(
        """
a=""
"""
    )
    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, Literal)
    assert stmt.value.value == ""


def test_empty_2():
    statements = parse_code(
        """
a=''
"""
    )
    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, Literal)
    assert stmt.value.value == ""


def test_string_format():
    statements = parse_code(
        """
a="j{{o}}s"
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, StringFormat)
    assert isinstance(stmt.value._variables[0][0], Reference)
    assert [x[0].name for x in stmt.value._variables] == ["o"]


def test_string_format_2():
    statements = parse_code(
        """
a="j{{c.d}}s"
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, StringFormat)
    assert len(stmt.value._variables) == 1
    assert len(stmt.value._variables[0]) == 2
    assert isinstance(stmt.value._variables[0][0], AttributeReference)
    assert stmt.value._variables[0][0].instance.name == "c"
    assert stmt.value._variables[0][0].attribute == "d"


def test_attribute_reference():
    statements = parse_code(
        """
a=a::b::c.d
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, AttributeReference)
    assert isinstance(stmt.value.instance, Reference)
    assert stmt.value.instance.full_name == "a::b::c"
    assert stmt.value.attribute == "d"


def test_is_defined():
    statements = parse_code(
        """
implement Test1 using tt when self.other is defined
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineImplement)
    assert isinstance(stmt.select, IsDefined)
    assert stmt.select.attr.name == "self"
    assert stmt.select.name == "other"


def test_is_defined_implicit_self():
    statements = parse_code(
        """
implement Test1 using tt when other is defined
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineImplement)
    assert isinstance(stmt.select, IsDefined)
    assert stmt.select.attr is None
    assert stmt.select.name == "other"


def test_is_defined_short():
    statements = parse_code(
        """
implement Test1 using tt when a.other is defined
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineImplement)
    assert isinstance(stmt.select, IsDefined)
    assert isinstance(stmt.select.attr, Reference)
    assert stmt.select.attr.name == "a"
    assert stmt.select.name == "other"


def assert_is_non_value(x):
    assert isinstance(x, Literal)
    assert isinstance(x.value, NoneValue)


def compare_attr(attr, name, mytype, defs, multi=False, opt=False):
    assert str(attr.name) == name
    defs(attr.default)
    assert attr.type.multi == multi
    assert str(attr.type.basetype) == mytype
    assert attr.type.nullable == opt


def assert_is_none(x):
    assert x is None


def assert_equals(x, y):
    assert x == y


def test_define_list_attribute():
    statements = parse_code(
        """
entity Jos:
  bool[] bar
  ip::ip[] ips = ["a"]
  string[] floom = []
  string[] floomx = ["a", "b"]
  string[]? floomopt = null
end"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineEntity)
    assert len(stmt.attributes) == 5

    compare_attr(stmt.attributes[0], "bar", "bool", assert_is_none, multi=True)
    compare_attr(stmt.attributes[2], "floom", "string", lambda x: assert_equals([], x.items), multi=True)

    def compare_default(list):
        def comp(x):
            assert len(list) == len(x.items)
            for one, it in zip(list, x.items):
                assert isinstance(it, Literal)
                assert it.value == one

        return comp

    compare_attr(stmt.attributes[1], "ips", "ip::ip", compare_default(["a"]), multi=True)
    compare_attr(stmt.attributes[3], "floomx", "string", compare_default(["a", "b"]), multi=True)
    compare_attr(stmt.attributes[4], "floomopt", "string", assert_is_non_value, opt=True, multi=True)


def test_define_dict_attribute():
    statements = parse_code(
        """
entity Jos:
  dict bar
  dict foo = {}
  dict blah = {"a":"a"}
  dict? xxx = {"a":"a"}
  dict? xxxx = null
end"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineEntity)
    assert len(stmt.attributes) == 5

    compare_attr(stmt.attributes[0], "bar", "dict", assert_is_none)
    compare_attr(stmt.attributes[1], "foo", "dict", lambda x: assert_equals([], x.items))

    def compare_default(list):
        def comp(x):
            assert len(list) == len(x.items)
            for (ok, ov), (k, v) in zip(list, x.items):
                assert k == ok
                assert ov == v.value

        return comp

    compare_attr(stmt.attributes[2], "blah", "dict", compare_default([("a", "a")]))
    compare_attr(stmt.attributes[3], "xxx", "dict", compare_default([("a", "a")]), opt=True)
    compare_attr(stmt.attributes[4], "xxxx", "dict", assert_is_non_value, opt=True)


def test_lexer():
    parse_code(
        """
#test
//test2
a=0.5
b=""
"""
    )


def test_eol_comment():
    parse_code(
        """a="a"
    # valid_target_types: tosca.capabilities.network.Bindable"""
    )


def test_mls():
    statements = parse_code(
        """
entity MANO:
    \"""
        This entity provides management, orchestration and monitoring

        More test
    \"""
end
"""
    )
    assert len(statements) == 1
    stmt = statements[0]

    assert isinstance(stmt, DefineEntity)

    mls = stmt.comment

    print(mls)

    assert (
        str(mls)
        == """
        This entity provides management, orchestration and monitoring

        More test
    """
    )

    assert (
        str(stmt.type.comment)
        == """
        This entity provides management, orchestration and monitoring

        More test
    """
    )


def test_bad():
    with pytest.raises(ParserException):
        parse_code(
            """
a = b.c
a=a::b::c.
"""
        )


def test_bad_2():
    with pytest.raises(ParserException):
        parse_code(
            """
a=|
"""
        )


def test_error_on_relation():
    with pytest.raises(ParserException) as e:
        parse_code(
            """
Host.provider [1] -- Provider test"""
        )
    assert e.value.location.file == "test"
    assert e.value.location.lnr == 3
    assert e.value.location.start_char == 2


def test_doc_string_on_new_relation():
    statements = parse_code(
        """
File.host [1] -- Host
\"""
Each file needs to be associated with a host
\"""
"""
    )
    assert len(statements) == 1

    stmt = statements[0]
    assert str(stmt.comment).strip() == "Each file needs to be associated with a host"


def test_doc_string_on_relation():
    statements = parse_code(
        """
File file [1] -- [0:] Host host
\"""
Each file needs to be associated with a host
\"""
"""
    )
    assert len(statements) == 1

    stmt = statements[0]
    assert str(stmt.comment).strip() == "Each file needs to be associated with a host"


def test_doc_string_on_typedef():
    statements = parse_code(
        """
typedef foo as string matching /^a+$/
\"""
    Foo is a stringtype that only allows "a"
\"""
"""
    )
    assert len(statements) == 1

    stmt = statements[0]
    assert str(stmt.comment).strip() == 'Foo is a stringtype that only allows "a"'


def test_doc_string_on_typedefault():
    statements = parse_code(
        """
typedef Foo as File(x=5)
\"""
    Foo is a stringtype that only allows "a"
\"""
"""
    )
    assert len(statements) == 1

    stmt = statements[0]
    assert str(stmt.comment).strip() == 'Foo is a stringtype that only allows "a"'


def test_doc_string_on_impl():
    statements = parse_code(
        """
implementation test for Host:
    \"""
        Bla bla
    \"""
end
"""
    )
    assert len(statements) == 1

    stmt = statements[0]
    assert str(stmt.comment).strip() == "Bla bla"
    assert str(stmt.type.comment).strip() == "Bla bla"


def test_doc_string_on_implements():
    statements = parse_code(
        """
implement Host using test
\"""
    Always use test!
\"""
\"""
    Not a comment
\"""

"""
    )
    assert len(statements) == 2

    stmt = statements[0]
    assert str(stmt.comment).strip() == "Always use test!"


def test_precise_lexer_positions():
    statements = parse_code(
        """
implement Test1 using tt when self.other is defined
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, DefineImplement)
    assert isinstance(stmt.select, IsDefined)
    assert stmt.select.attr.name == "self"
    assert str(stmt.select.name) == "other"


def test_list_extend_bad():
    with pytest.raises(ParserException):
        parse_code(
            """
    a+=b
    """
        )


def test_list_extend_good():
    statements = parse_code(
        """
z.a+=b
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, SetAttribute)
    assert stmt.list_only is True
    assert isinstance(stmt.value, Reference)
    assert stmt.value.name == "b"


def test_mapref():
    """Test extending entities
    """
    statements = parse_code(
        """
a = b.c["test"]
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, MapLookup)
    assert isinstance(stmt.value.themap, AttributeReference)
    assert stmt.value.themap.instance.name == "b"
    assert stmt.value.themap.attribute == "c"
    assert isinstance(stmt.value.key, Literal)
    assert stmt.value.key.value == "test"


def test_mapref_2():
    """Test extending entities
    """
    statements = parse_code(
        """
a = c["test"]
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, MapLookup)
    assert isinstance(stmt.value.themap, Reference)
    assert stmt.value.themap.name == "c"
    assert isinstance(stmt.value.key, Literal)
    assert stmt.value.key.value == "test"


def test_map_multi_ref():
    """Test extending entities
    """
    statements = parse_code(
        """
a = c["test"]["xx"]
"""
    )

    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, Assign)
    assert isinstance(stmt.value, MapLookup)
    assert isinstance(stmt.value.themap, MapLookup)
    assert isinstance(stmt.value.themap.themap, Reference)
    assert stmt.value.themap.themap.name == "c"
    assert isinstance(stmt.value.key, Literal)
    assert stmt.value.key.value == "xx"
    assert isinstance(stmt.value.themap.key, Literal)
    assert stmt.value.themap.key.value == "test"


def test_if_statement():
    """Test for the if statement
    """
    statements = parse_code(
        """
if test.field == "value":
    test.other = "otherValue"
end
"""
    )
    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, If)
    assert isinstance(stmt.condition, ExpressionStatement)
    assert isinstance(stmt.if_branch, BasicBlock)
    assert len(stmt.if_branch.get_stmts()) == 1
    assert isinstance(stmt.else_branch, BasicBlock)
    assert len(stmt.else_branch.get_stmts()) == 0


def test_if_else():
    """Test for the if statement with an else clause
    """
    statements = parse_code(
        """
if test.field == "value":
    test.other = "otherValue"
else:
    test.other = "altValue"
end
"""
    )
    assert len(statements) == 1
    stmt = statements[0]
    assert isinstance(stmt, If)
    assert isinstance(stmt.condition, ExpressionStatement)
    assert isinstance(stmt.if_branch, BasicBlock)
    assert len(stmt.if_branch.get_stmts()) == 1
    assert isinstance(stmt.else_branch, BasicBlock)
    assert len(stmt.else_branch.get_stmts()) == 1


def test_bool_str():
    """Test to string of bool literal renders inmanta true/false and not python
    """
    statements = parse_code(
        """
val1 = true
val2 = false
"""
    )
    assert len(statements) == 2
    assert isinstance(statements[0], Assign)
    assert isinstance(statements[1], Assign)
    assert str(statements[0].rhs) == "true"
    assert str(statements[1].rhs) == "false"


def test_1341_syntax_error_output_1():
    """
    Test the readability of a syntax error message.
    """
    with pytest.raises(ParserException) as pytest_e:
        parse_code(
            """
var=é,
)
            """
        )
    exc: ParserException = pytest_e.value
    assert exc.location.file == "test"
    assert exc.location.lnr == 2
    assert exc.location.start_char == 5
    assert exc.location.end_lnr == 2
    assert exc.location.end_char == 6
    assert exc.value == "é"
    assert exc.msg == "Syntax error: Illegal character 'é'"


def test_1341_syntax_error_output_2():
    """
    Test the readability of a syntax error message.
    """
    with pytest.raises(ParserException) as pytest_e:
        parse_code(
            """
deployment2 = k8s::Deployment(
    name="hello-nginx2",
    var=🤔,
    cluster=cluster
)
            """
        )
    exc: ParserException = pytest_e.value
    assert exc.location.file == "test"
    assert exc.location.lnr == 4
    assert exc.location.start_char == 9
    assert exc.location.end_lnr == 4
    assert exc.location.end_char == 10
    assert exc.value == "🤔"
    assert exc.msg == "Syntax error: Illegal character '🤔'"


def test_1341_syntax_error_output_3():
    """
    Test the readability of a syntax error message.
    """
    with pytest.raises(ParserException) as pytest_e:
        parse_code(
            """
é
            """
        )
    exc: ParserException = pytest_e.value
    assert exc.location.file == "test"
    assert exc.location.lnr == 2
    assert exc.location.start_char == 1
    assert exc.location.end_lnr == 2
    assert exc.location.end_char == 2
    assert exc.value == "é"
    assert exc.msg == "Syntax error: Illegal character 'é'"


def test_1341_syntax_error_output_4():
    """
    Test the readability of a syntax error message.
    """
    with pytest.raises(ParserException) as pytest_e:
        parse_code(
            """
aé=66
            """
        )
    exc: ParserException = pytest_e.value
    assert exc.location.file == "test"
    assert exc.location.lnr == 2
    assert exc.location.start_char == 2
    assert exc.location.end_lnr == 2
    assert exc.location.end_char == 3
    assert exc.value == "é"
    assert exc.msg == "Syntax error: Illegal character 'é'"


def test_1341_syntax_error_output_5():
    """
    Test the readability of a syntax error message.
    """
    with pytest.raises(ParserException) as pytest_e:
        parse_code(
            """
K8éééYamlResource.cluster [1] -- Cluster
            """
        )
    exc: ParserException = pytest_e.value
    assert exc.location.file == "test"
    assert exc.location.lnr == 2
    assert exc.location.start_char == 3
    assert exc.location.end_lnr == 2
    assert exc.location.end_char == 4
    assert exc.value == "é"
    assert exc.msg == "Syntax error: Illegal character 'é'"


def test_640_syntax_error_output_6():
    """
    Test the readability of a syntax error message.
    """
    with pytest.raises(ParserException) as pytest_e:
        parse_code(
            """
typedef positive as number matching self >= 1-
            """
        )
    exc: ParserException = pytest_e.value
    assert exc.location.file == "test"
    assert exc.location.lnr == 2
    assert exc.location.start_char == 46
    assert exc.location.end_lnr == 2
    assert exc.location.end_char == 47
    assert exc.value == "-"
    assert exc.msg == "Syntax error: Illegal character '-'"


def test_1766_empty_model_single_newline():
    statements = parse_code(
        """
        """
    )
    assert len(statements) == 0


def test_1766_empty_model_multiple_newline():
    statements = parse_code(
        """



        """
    )
    assert len(statements) == 0


def test_1707_out_of_place_regex():
    with pytest.raises(ParserException) as pytest_e:
        parse_code(
            """
/some_out_of_place_regex/
            """,
        )
    exc: ParserException = pytest_e.value
    assert exc.msg == "Syntax error at token /some_out_of_place_regex/"


def test_multiline_string_interpolation():
    statements = parse_code(
        """
str = \"\"\"
    var == {{var}}
\"\"\"
        """,
    )
    assert len(statements) == 1
    assert isinstance(statements[0], Assign)
    assert isinstance(statements[0].rhs, StringFormat)


def test_1804_bool_condition_as_bool():
    statements = parse_code(
        """
if false and true == true:
end
        """,
    )
    assert len(statements) == 1
    if_stmt = statements[0]
    assert isinstance(if_stmt, If)
    and_stmt = if_stmt.condition
    assert isinstance(and_stmt, And)
    assert len(and_stmt.children) == 2
    false_stmt = and_stmt.children[0]
    assert isinstance(false_stmt, Literal)
    assert isinstance(false_stmt.value, bool)
    assert false_stmt.value is False


def test_1573_condition_dict_lookup():
    statements = parse_code(
        """
dct = {"b": true}

if dct["b"]:
end
        """,
    )
    assert len(statements) == 2
    assert isinstance(statements[1], If)


@pytest.mark.parametrize(
    "expression,expected_tree",
    [
        ("42 == 42 and not false", (And, [(Equals, [(Literal, 42), (Literal, 42)]), (Not, [(Literal, False)])])),
        (
            "42 in [12, 42] or 'test' in []",
            (
                Or,
                [
                    (In, [(Literal, 42), (CreateList, [(Literal, 12), (Literal, 42)])]),
                    (In, [(Literal, "test"), (CreateList, [])]),
                ],
            ),
        ),
        ("not (42 in x)", (Not, [(In, [(Literal, 42), (Reference, "x")])])),
        ("not 42 in x", (Not, [(In, [(Literal, 42), (Reference, "x")])])),
        ("x or y.u is defined", (Or, [(Reference, "x"), (IsDefined, [(Reference, "y")])])),
    ],
)
def test_1815_conditional_expressions(expression, expected_tree):
    statements = parse_code(
        f"""
__x__ = {expression}
        """,
    )
    assert len(statements) == 1
    assign_stmt = statements[0]
    assert isinstance(assign_stmt, Assign)

    def expression_asserter(expression: ExpressionStatement, expected_tree):
        assert isinstance(expression, expected_tree[0])
        if isinstance(expression, Literal):
            assert expression.value == expected_tree[1]
        elif isinstance(expression, Reference):
            assert expression.name == expected_tree[1]
        elif isinstance(expression, ReferenceStatement):
            assert len(expression.children) == len(expected_tree[1])
            for child, child_expected in zip(expression.children, expected_tree[1]):
                if child_expected is None:
                    continue
                expression_asserter(child, child_expected)
        else:
            raise Exception("this test does not support %s" % type(expression))

    expression_asserter(assign_stmt.rhs, expected_tree)


def test_relation_deprecated_syntax():
    with warnings.catch_warnings(record=True) as w:
        parse_code(
            """
entity A:
end

entity B:
end

A aa [1] -- [0:] B bb
            """,
        )
        assert len(w) == 1
        assert issubclass(w[0].category, SyntaxDeprecationWarning)
        assert str(w[0].message) == (
            "The relation definition syntax `A aa [1:1] -- [0:] B bb` is deprecated."
            " Please use `A.bb [0:] -- B.aa [1:1]` instead. (test:8)"
        )



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import asyncio
import datetime
import logging
import os
import shutil
import subprocess
import uuid
from asyncio import Semaphore
from typing import List

import pytest

from inmanta import config, data
from inmanta.const import ParameterSource
from inmanta.data import Compile, Report
from inmanta.deploy import cfg_env
from inmanta.server import SLICE_COMPILER, SLICE_SERVER
from inmanta.server import config as server_config
from inmanta.server import protocol
from inmanta.server.bootloader import InmantaBootloader
from inmanta.server.protocol import Server
from inmanta.server.services.compilerservice import CompilerService, CompileRun, CompileStateListener
from inmanta.util import ensure_directory_exist
from utils import LogSequence, report_db_index_usage, retry_limited, wait_for_version

logger = logging.getLogger("inmanta.test.server.compilerservice")


@pytest.fixture
async def compilerservice(server_config, init_dataclasses_and_load_schema):
    server = Server()
    cs = CompilerService()
    await cs.prestart(server)
    await cs.start()
    yield cs
    await cs.prestop()
    await cs.stop()


@pytest.mark.asyncio
async def test_scheduler(server_config, init_dataclasses_and_load_schema, caplog):
    """Test the scheduler part in isolation, mock out compile runner and listen to state updates"""

    class Collector(CompileStateListener):
        """
            Collect all state updates, optionally hang the processing of listeners
        """

        def __init__(self):
            self.seen = []
            self.preseen = []
            self.lock = Semaphore(1)

        def reset(self):
            self.seen = []
            self.preseen = []

        async def compile_done(self, compile: data.Compile):
            self.preseen.append(compile)
            print("Got compile done for ", compile.remote_id)
            async with self.lock:
                self.seen.append(compile)

        async def hang(self):
            await self.lock.acquire()

        def release(self):
            self.lock.release()

        def verify(self, envs: uuid.UUID):
            assert sorted([x.remote_id for x in self.seen]) == sorted(envs)
            self.reset()

    class HangRunner(object):
        """
            compile runner mock, hang until released
        """

        def __init__(self):
            self.lock = Semaphore(0)
            self.started = False
            self.done = False
            self.version = None

        async def run(self):
            self.started = True
            await self.lock.acquire()
            self.done = True
            return True

        def release(self):
            self.lock.release()

    class HookedCompilerService(CompilerService):
        """
            hook in the hangrunner
        """

        def __init__(self):
            super(HookedCompilerService, self).__init__()
            self.locks = {}

        def _get_compile_runner(self, compile: data.Compile, project_dir: str):
            print("Get Run: ", compile.remote_id, compile.id)
            runner = HangRunner()
            self.locks[compile.remote_id] = runner
            return runner

        def get_runner(self, remote_id: uuid.UUID) -> HangRunner:
            return self.locks.get(remote_id)

    # manual setup of server
    server = Server()
    cs = HookedCompilerService()
    await cs.prestart(server)
    await cs.start()
    collector = Collector()
    cs.add_listener(collector)

    async def request_compile(env: data.Environment) -> uuid.UUID:
        """Request compile for given env, return remote_id"""
        u1 = uuid.uuid4()
        await cs.request_recompile(env, False, False, u1)
        results = await data.Compile.get_by_remote_id(env.id, u1)
        assert len(results) == 1
        assert results[0].remote_id == u1
        print("request: ", u1, results[0].id)
        return u1

    # setup projects in the database
    project = data.Project(name="test")
    await project.insert()
    env1 = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env1.insert()
    env2 = data.Environment(name="dev2", project=project.id, repo_url="", repo_branch="")
    await env2.insert()

    # setup series of compiles for two envs
    # e1 is for a plain run
    # e2 is for server restart
    e1 = [await request_compile(env1) for i in range(3)]
    e2 = [await request_compile(env2) for i in range(4)]
    print("env 1:", e1)

    async def check_compile_in_sequence(env: data.Environment, remote_ids: List[uuid.UUID], idx: int):
        """
        Check integrity of a compile sequence and progress the hangrunner.
        """
        before = remote_ids[:idx]

        for rid in before:
            prevrunner = cs.get_runner(rid)
            assert prevrunner.done

        if idx < len(remote_ids):
            current = remote_ids[idx]
            after = remote_ids[idx + 1 :]

            assert await cs.is_compiling(env.id) == 200

            await retry_limited(lambda: cs.get_runner(current) is not None, 1)
            await retry_limited(lambda: cs.get_runner(current).started, 1)

            for rid in after:
                nextrunner = cs.get_runner(rid)
                assert nextrunner is None

            cs.get_runner(current).release()
            await asyncio.sleep(0)
            await retry_limited(lambda: cs.get_runner(current).done, 1)

        else:

            async def isdone():
                return await cs.is_compiling(env.id) == 204

            await retry_limited(isdone, 1)

    # run through env1, entire sequence
    for i in range(4):
        await check_compile_in_sequence(env1, e1, i)
    collector.verify(e1)
    print("env1 done")

    print("env2 ", e2)
    # make event collector hang
    await collector.hang()
    # progress two steps into env2
    for i in range(2):
        await check_compile_in_sequence(env2, e2, i)

    assert not collector.seen
    print(collector.preseen)
    await retry_limited(lambda: len(collector.preseen) == 2, 1)

    # test server restart
    await cs.prestop()
    await cs.stop()

    # in the log, find cancel of compile(hangs) and handler(hangs)
    LogSequence(caplog, allow_errors=False).contains("inmanta.util", logging.WARNING, "was cancelled").contains(
        "inmanta.util", logging.WARNING, "was cancelled"
    ).no_more_errors()

    print("restarting")

    # restart new server
    cs = HookedCompilerService()
    await cs.prestart(server)
    await cs.start()
    collector = Collector()
    cs.add_listener(collector)

    # complete the sequence, expect re-run of third compile
    for i in range(3):
        print(i)
        await check_compile_in_sequence(env2, e2[2:], i)

    # all are re-run, entire sequence present
    collector.verify(e2)

    await report_db_index_usage()


@pytest.mark.asyncio
async def test_compile_runner(server, tmpdir, client):
    # create project, printing env
    project_source_dir = os.path.join(tmpdir, "src")

    project_template = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "data", "project")

    shutil.copytree(project_template, project_source_dir)
    subprocess.check_output(["git", "init"], cwd=project_source_dir)
    subprocess.check_output(["git", "add", "*"], cwd=project_source_dir)
    subprocess.check_output(["git", "config", "user.name", "Unit"], cwd=project_source_dir)
    subprocess.check_output(["git", "config", "user.email", "unit@test.example"], cwd=project_source_dir)
    subprocess.check_output(["git", "commit", "-m", "unit test"], cwd=project_source_dir)

    testmarker_env = "TESTMARKER"
    no_marker = "__no__marker__"
    marker_print = "_INM_MM:"
    marker_print2 = "_INM_MM2:"
    marker_print3 = "_INM_MM3:"

    def make_main(marker_print):
        # add main.cf
        with open(os.path.join(project_source_dir, "main.cf"), "w", encoding="utf-8") as fd:
            fd.write(
                f"""
    marker = std::get_env("{testmarker_env}","{no_marker}")
    std::print("{marker_print} {{{{marker}}}}")

    host = std::Host(name="test", os=std::linux)
    std::ConfigFile(host=host, path="/etc/motd", content="1234")
    """
            )

        subprocess.check_output(["git", "add", "main.cf"], cwd=project_source_dir)
        subprocess.check_output(["git", "commit", "-m", "unit test 2"], cwd=project_source_dir)

    make_main(marker_print)
    subprocess.check_output("git checkout -b alt".split(), cwd=project_source_dir)
    make_main(marker_print2)

    project_work_dir = os.path.join(tmpdir, "work")
    ensure_directory_exist(project_work_dir)

    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url=project_source_dir, repo_branch="master")
    await env.insert()

    env2 = data.Environment(name="devalt", project=project.id, repo_url=project_source_dir, repo_branch="alt")
    await env2.insert()

    async def compile_and_assert(env, export=True, meta={}, env_vars={}, update=False):
        compile = data.Compile(
            remote_id=uuid.uuid4(),
            environment=env.id,
            do_export=export,
            metadata=meta,
            environment_variables=env_vars,
            force_update=update,
        )
        await compile.insert()

        # compile with export
        cr = CompileRun(compile, project_work_dir)
        await cr.run()

        # get and process reports
        result = await client.get_report(compile.id)
        assert result.code == 200
        stages = result.result["report"]["reports"]
        if export:
            assert cr.version > 0
            result = await client.get_version(env.id, cr.version)
            assert result.code == 200
            metadata = result.result["model"]["version_info"]["export_metadata"]
            for k, v in meta.items():
                assert k in metadata
                assert v == metadata[k]

        print(stages)

        stage_by_name = {stage["name"]: stage for stage in stages}

        return cr, stage_by_name

    # with export
    compile, stages = await compile_and_assert(env, True, meta={"type": "Test"})
    assert stages["Init"]["returncode"] == 0
    assert stages["Cloning repository"]["returncode"] == 0
    assert stages["Recompiling configuration model"]["returncode"] == 0
    out = stages["Recompiling configuration model"]["outstream"]
    assert f"{marker_print} {no_marker}" in out
    assert len(stages) == 3
    assert compile.version is not None

    # no export
    compile, stages = await compile_and_assert(env, False)
    assert stages["Init"]["returncode"] == 0
    assert stages["Recompiling configuration model"]["returncode"] == 0
    out = stages["Recompiling configuration model"]["outstream"]
    assert f"{marker_print} {no_marker}" in out
    assert len(stages) == 2
    assert compile.version is None

    # env vars
    marker = str(uuid.uuid4())
    compile, stages = await compile_and_assert(env, False, env_vars={testmarker_env: marker})
    assert len(compile.request.environment_variables) == 1
    assert stages["Init"]["returncode"] == 0
    assert f"Using extra environment variables during compile TESTMARKER='{marker}'" in stages["Init"]["outstream"]
    assert stages["Recompiling configuration model"]["returncode"] == 0
    out = stages["Recompiling configuration model"]["outstream"]
    assert f"{marker_print} {marker}" in out
    assert len(stages) == 2
    assert compile.version is None

    # switch branch
    compile, stages = await compile_and_assert(env2, False)
    assert stages["Init"]["returncode"] == 0
    assert stages["switching branch from master to alt"]["returncode"] == 0
    assert stages["Recompiling configuration model"]["returncode"] == 0
    out = stages["Recompiling configuration model"]["outstream"]
    assert f"{marker_print2} {no_marker}" in out
    assert len(stages) == 3
    assert compile.version is None

    # update with no update
    compile, stages = await compile_and_assert(env2, False, update=True)
    assert stages["Init"]["returncode"] == 0
    assert stages["Fetching changes"]["returncode"] == 0
    assert stages["Pulling updates"]["returncode"] == 0
    assert stages["Updating modules"]["returncode"] == 0
    assert stages["Recompiling configuration model"]["returncode"] == 0
    out = stages["Recompiling configuration model"]["outstream"]
    assert f"{marker_print2} {no_marker}" in out
    assert len(stages) == 5
    assert compile.version is None

    make_main(marker_print3)
    compile, stages = await compile_and_assert(env2, False, update=True)
    assert stages["Init"]["returncode"] == 0
    assert stages["Fetching changes"]["returncode"] == 0
    assert stages["Pulling updates"]["returncode"] == 0
    assert stages["Updating modules"]["returncode"] == 0
    assert stages["Recompiling configuration model"]["returncode"] == 0
    out = stages["Recompiling configuration model"]["outstream"]
    assert f"{marker_print3} {no_marker}" in out
    assert len(stages) == 5
    assert compile.version is None


@pytest.mark.asyncio
async def test_e2e_recompile_failure(compilerservice: CompilerService):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    u1 = uuid.uuid4()
    await compilerservice.request_recompile(env, False, False, u1)
    u2 = uuid.uuid4()
    await compilerservice.request_recompile(env, False, False, u2)

    assert await compilerservice.is_compiling(env.id) == 200

    async def compile_done():
        res = await compilerservice.is_compiling(env.id)
        print(res)
        return res == 204

    await retry_limited(compile_done, 10)

    _, all_compiles = await compilerservice.get_reports(env)
    all_reports = {i["remote_id"]: await compilerservice.get_report(i["id"]) for i in all_compiles["reports"]}

    def assert_report(uid):
        code, report = all_reports[uid]
        report = report["report"]
        assert report["remote_id"] == uid
        assert not report["success"]
        reports = report["reports"]
        reports = {r["name"]: r for r in reports}

        # stages
        init = reports["Init"]
        assert not init["errstream"]
        assert "project found in" in init["outstream"] and "and no repository set" in init["outstream"]

        # compile
        comp = reports["Recompiling configuration model"]
        assert "Unable to find an inmanta project (project.yml expected)" in comp["errstream"]
        assert comp["returncode"] == 1
        return report["requested"], report["started"], report["completed"]

    r1, s1, f1 = assert_report(u1)
    r2, s2, f2 = assert_report(u2)

    assert r2 > r1
    assert r1 < s1 < f1
    assert r2 < s2 < f2
    assert f1 < s2


@pytest.mark.asyncio(timeout=90)
async def test_server_recompile(server, client, environment, monkeypatch):
    """
        Test a recompile on the server and verify recompile triggers
    """
    config.Config.set("server", "auto-recompile-wait", "0")

    project_dir = os.path.join(server.get_slice(SLICE_SERVER)._server_storage["environments"], str(environment))
    project_source = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "data", "project")
    print("Project at: ", project_dir)

    shutil.copytree(project_source, project_dir)
    subprocess.check_output(["git", "init"], cwd=project_dir)
    subprocess.check_output(["git", "add", "*"], cwd=project_dir)
    subprocess.check_output(["git", "config", "user.name", "Unit"], cwd=project_dir)
    subprocess.check_output(["git", "config", "user.email", "unit@test.example"], cwd=project_dir)
    subprocess.check_output(["git", "commit", "-m", "unit test"], cwd=project_dir)

    # Set environment variable to be passed to the compiler
    key_env_var = "TEST_MESSAGE"
    value_env_var = "a_message"
    monkeypatch.setenv(key_env_var, value_env_var)

    # add main.cf
    with open(os.path.join(project_dir, "main.cf"), "w", encoding="utf-8") as fd:
        fd.write(
            f"""
        host = std::Host(name="test", os=std::linux)
        std::ConfigFile(host=host, path="/etc/motd", content="1234")
        std::print(std::get_env("{key_env_var}"))
"""
        )

    logger.info("request a compile")
    result = await client.notify_change(environment)
    assert result.code == 200

    logger.info("wait for 1")
    versions = await wait_for_version(client, environment, 1)
    assert versions["versions"][0]["total"] == 1
    assert versions["versions"][0]["version_info"]["export_metadata"]["type"] == "api"

    # get compile reports and make sure the environment variables are not logged
    reports = await client.get_reports(environment)
    assert reports.code == 200
    assert len(reports.result["reports"]) == 1
    env_vars_compile = reports.result["reports"][0]["environment_variables"]
    assert key_env_var not in env_vars_compile

    # get report
    compile_report = await client.get_report(reports.result["reports"][0]["id"])
    assert compile_report.code == 200
    report_map = {r["name"]: r for r in compile_report.result["report"]["reports"]}
    assert value_env_var in report_map["Recompiling configuration model"]["outstream"]

    # set a parameter without requesting a recompile
    await client.set_param(environment, id="param1", value="test", source=ParameterSource.plugin)
    versions = await wait_for_version(client, environment, 1)
    assert versions["count"] == 1

    logger.info("request second compile")
    # set a new parameter and request a recompile
    await client.set_param(environment, id="param2", value="test", source=ParameterSource.plugin, recompile=True)
    logger.info("wait for 2")
    versions = await wait_for_version(client, environment, 2)
    assert versions["versions"][0]["version_info"]["export_metadata"]["type"] == "param"
    assert versions["count"] == 2

    # update the parameter to the same value -> no compile
    await client.set_param(environment, id="param2", value="test", source=ParameterSource.plugin, recompile=True)
    versions = await wait_for_version(client, environment, 2)
    assert versions["count"] == 2

    # update the parameter to a new value
    await client.set_param(environment, id="param2", value="test2", source=ParameterSource.plugin, recompile=True)
    versions = await wait_for_version(client, environment, 3)
    logger.info("wait for 3")
    assert versions["count"] == 3

    # clear the environment
    state_dir = server_config.state_dir.get()
    project_dir = os.path.join(state_dir, "server", "environments", environment)
    assert os.path.exists(project_dir)

    await client.clear_environment(environment)

    assert not os.path.exists(project_dir)


@pytest.mark.asyncio(timeout=90)
async def test_compileservice_queue(mocked_compiler_service_block, server, client, environment):
    """
        Test the inspection of the compile queue. The compile runner is mocked out so the "started" field does not have the
        correct value in this test.
    """
    env = await data.Environment.get_by_id(environment)
    config.Config.set("server", "auto-recompile-wait", "0")
    compilerslice: CompilerService = server.get_slice(SLICE_COMPILER)

    result = await client.get_compile_queue(environment)
    assert len(result.result["queue"]) == 0
    assert result.code == 200

    # request a compile
    compile_id1 = uuid.uuid4()
    await compilerslice.request_recompile(env=env, force_update=False, do_export=False, remote_id=compile_id1)

    # api should return one
    result = await client.get_compile_queue(environment)
    assert len(result.result["queue"]) == 1
    assert result.result["queue"][0]["remote_id"] == str(compile_id1)
    assert result.code == 200

    # request a compile
    compile_id2 = uuid.uuid4()
    await compilerslice.request_recompile(env=env, force_update=False, do_export=False, remote_id=compile_id2)

    # api should return two
    result = await client.get_compile_queue(environment)
    assert len(result.result["queue"]) == 2
    assert result.result["queue"][1]["remote_id"] == str(compile_id2)
    assert result.code == 200

    # finish a compile and wait for service to take on next
    current_task = compilerslice._recompiles[env.id]

    run = mocked_compiler_service_block.get()
    run.block = False

    while current_task is compilerslice._recompiles[env.id]:
        await asyncio.sleep(0.01)

    # api should return one when ready
    result = await client.get_compile_queue(environment)
    assert len(result.result["queue"]) == 1
    assert result.code == 200

    # finish a compile
    current_task = compilerslice._recompiles[env.id]

    run = mocked_compiler_service_block.get()
    run.block = False

    while env.id in compilerslice._recompiles and current_task is compilerslice._recompiles[env.id]:
        await asyncio.sleep(0.01)

    # return

    # api should return none
    result = await client.get_compile_queue(environment)
    assert len(result.result["queue"]) == 0
    assert result.code == 200


@pytest.fixture(scope="function")
async def server_with_frequent_cleanups(server_pre_start, server_config, async_finalizer):
    config.Config.set("server", "compiler-report-retention", "2")
    config.Config.set("server", "cleanup-compiler-reports_interval", "1")
    ibl = InmantaBootloader()
    await ibl.start()
    yield ibl.restserver
    await asyncio.wait_for(ibl.stop(), 15)


@pytest.fixture(scope="function")
def client_for_cleanup(server_with_frequent_cleanups):
    client = protocol.Client("client")
    yield client


@pytest.fixture(scope="function")
async def environment_for_cleanup(client_for_cleanup, server_with_frequent_cleanups):
    """
        Create a project and environment. This fixture returns the uuid of the environment
    """
    result = await client_for_cleanup.create_project("env-test")
    assert result.code == 200
    project_id = result.result["project"]["id"]

    result = await client_for_cleanup.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    cfg_env.set(env_id)

    yield env_id


@pytest.fixture
async def old_compile_report(server_with_frequent_cleanups, environment_for_cleanup):
    now = datetime.datetime.now()
    time_of_compile = now - datetime.timedelta(days=30)
    compile_id = uuid.UUID("c00cc33f-f70f-4800-ad01-ff042f67118f")
    old_compile = {
        "id": compile_id,
        "remote_id": uuid.UUID("c9a10da1-9bf6-4152-8461-98adc02c4cee"),
        "environment": uuid.UUID(environment_for_cleanup),
        "requested": time_of_compile,
        "started": time_of_compile,
        "completed": time_of_compile,
        "do_export": True,
        "force_update": True,
        "metadata": {"type": "api", "message": "Recompile trigger through API call"},
        "environment_variables": {},
        "success": True,
        "handled": True,
        "version": 1,
    }
    await Compile(**old_compile).insert()
    new_compile = {**old_compile, "id": uuid.uuid4(), "requested": now, "started": now, "completed": now}
    await Compile(**new_compile).insert()

    report1 = {
        "id": uuid.UUID("2baa0175-9169-40c5-a546-64d646f62da6"),
        "started": time_of_compile,
        "completed": time_of_compile,
        "command": "",
        "name": "Init",
        "errstream": "",
        "outstream": "Using extra environment variables during compile \n",
        "returncode": 0,
        "compile": compile_id,
    }
    report2 = {
        **report1,
        "id": uuid.UUID("2a86d2a0-666f-4cca-b0a8-13e0379128d5"),
        "command": "python -m inmanta.app -vvv export -X",
        "name": "Recompiling configuration model",
    }
    await Report(**report1).insert()
    await Report(**report2).insert()
    yield compile_id


@pytest.mark.asyncio
async def test_compileservice_cleanup(
    server_with_frequent_cleanups, client_for_cleanup, environment_for_cleanup, old_compile_report
):
    # There is a new and an older report in the database
    result = await client_for_cleanup.get_reports(environment_for_cleanup)
    assert result.code == 200
    assert len(result.result["reports"]) == 2

    result = await client_for_cleanup.get_report(old_compile_report)
    assert result.code == 200
    assert len(result.result["report"]["reports"]) > 0

    compilerslice: CompilerService = server_with_frequent_cleanups.get_slice(SLICE_COMPILER)
    await compilerslice._cleanup()

    # The old report is deleted after cleanup
    result = await client_for_cleanup.get_report(old_compile_report)
    assert result.code == 404
    reports_after_cleanup = await Report.get_list(compile=old_compile_report)
    assert len(reports_after_cleanup) == 0

    # The new report is still there
    result = await client_for_cleanup.get_reports(environment_for_cleanup)
    assert result.code == 200
    assert len(result.result["reports"]) == 1


@pytest.mark.slowtest
@pytest.mark.asyncio
async def test_compileservice_cleanup_on_trigger(client_for_cleanup, environment_for_cleanup, old_compile_report):
    # Two reports are in the table
    result = await client_for_cleanup.get_reports(environment_for_cleanup)
    assert result.code == 200
    assert len(result.result["reports"]) == 2

    result = await client_for_cleanup.get_report(old_compile_report)
    assert result.code == 200
    assert len(result.result["report"]["reports"]) > 0

    await asyncio.sleep(3)

    # Both reports should be deleted after the triggered cleanup
    result = await client_for_cleanup.get_reports(environment_for_cleanup)
    assert result.code == 200
    assert len(result.result["reports"]) == 0
